from refinitiv.data.delivery._stream.contrib._response import (
    NullContribResponse,
    ErrorContribResponse,
    AckContribResponse,
)


def test_null_contrib_response():
    response = NullContribResponse()

    assert response.ack_id == None
    assert response.debug == {}
    assert response.error == ""
    assert response.is_success is False
    assert response.nak_code == ""
    assert response.nak_message == ""
    assert response.type == ""
    assert (
        "<refinitiv.data.delivery._stream.contrib._response.NullContribResponse object at "
        in repr(response)
    )


def test_error_contrib_response():
    input_debug = {
        "File": "/local/jsonToRwfSimple.C",
        "Line": 7243,
        "Offset": 233,
        "Message": '{"Ack": true, "ID": 2, "Message": {"Fields": {"ASK": 1.2, "BID": 0.86, "PRIMACT_1": 103, "SEC_ACT_1": 203}, "ID": 0, "Type": "Update", "Domain": "MarketPrice"}, "PostID": 11, "Type": "Post", "Key": {"Name": "TEST02=NIK", "Service": "DDS_TRCE"}, "Domain": "MarketPrice"}',
    }
    expected_debug = input_debug
    input_text = "JSON Unexpected Value. Received 'DDS_TRCE' for key 'Service'"
    expected_text = input_text
    response = ErrorContribResponse(
        {
            "ID": 2,
            "Type": "Error",
            "NakCode": "NakCode",
            "Text": input_text,
            "Debug": input_debug,
        }
    )

    assert response.ack_id == None
    assert response.debug == expected_debug
    assert response.error == expected_text
    assert response.is_success is False
    assert response.nak_code == "NakCode"
    assert response.nak_message == ""
    assert response.type == "Error"
    assert repr(response) == str(
        {
            "Type": "Error",
            "Text": "",
            "Debug": {
                "File": "/local/jsonToRwfSimple.C",
                "Line": 7243,
                "Offset": 233,
                "Message": '{"Ack": true, "ID": 2, "Message": {"Fields": {"ASK": 1.2, "BID": 0.86, "PRIMACT_1": 103, "SEC_ACT_1": 203}, "ID": 0, "Type": "Update", "Domain": "MarketPrice"}, "PostID": 11, "Type": "Post", "Key": {"Name": "TEST02=NIK", "Service": "DDS_TRCE"}, "Domain": "MarketPrice"}',
            },
        }
    )


def test_ack_contrib_response():
    input_text = "[1]: Contribution Accepted"
    expected_text = input_text

    input_ackid = 5
    expected_ackid = 5
    response = AckContribResponse(
        {
            "ID": 2,
            "Type": "Ack",
            "AckID": input_ackid,
            "Text": input_text,
            "Key": {"Service": "ATS_GLOBAL_1", "Name": "TEST"},
        }
    )

    assert response.ack_id == expected_ackid
    assert response.debug == {}
    assert response.error == ""
    assert response.is_success is True
    assert response.nak_code == ""
    assert response.nak_message == expected_text
    assert response.type == "Ack"
    assert repr(response) == str({"Type": "Ack", "AckId": 5})


def test_ack_contrib_response_with_nak_code():
    input_text = "[1]: Contribution Accepted"
    expected_text = input_text

    input_ackid = 5
    expected_ackid = 5
    response = AckContribResponse(
        {
            "ID": 2,
            "Type": "Ack",
            "AckID": input_ackid,
            "Text": input_text,
            "Key": {"Service": "ATS_GLOBAL_1", "Name": "TEST"},
            "NakCode": "NakCode",
        }
    )

    assert response.ack_id == expected_ackid
    assert response.debug == {}
    assert response.error == ""
    assert response.is_success is False
    assert response.nak_code == "NakCode"
    assert response.nak_message == expected_text
    assert response.type == "Ack"
    assert repr(response) == str(
        {
            "Type": "Ack",
            "AckId": 5,
            "NakCode": "NakCode",
            "Message": "[1]: Contribution Accepted",
        }
    )



from datetime import timedelta
from typing import Callable, Optional, TYPE_CHECKING

from ._news_data_provider_layer import NewsDataProviderLayer
from ._sort_order import SortOrder
from ..._content_type import ContentType
from ..._tools import create_repr

if TYPE_CHECKING:
    from ..._core.session import Session
    from ..._types import ExtendedParams, OptDateTime


class Definition(NewsDataProviderLayer):
    """
    This class describes parameters to retrieve data for news headlines.

    Parameters
    ----------
    query: str
        The user search query.

    count: int, optional
        Count to limit number of headlines. Min value is 0. Default: 10

    date_from: str or timedelta, optional
        Beginning of date range.
        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    date_to: str or timedelta, optional
        End of date range.
        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    sort_order: SortOrder
        Value from SortOrder enum. Default: SortOrder.new_to_old

    extended_params: dict, optional
        Other parameters can be provided if necessary

    Examples
    --------
    >>> from refinitiv.data.content import news
    >>> definition = news.headlines.Definition(
    ...     "Refinitiv",
    ...     date_from="20.03.2021",
    ...     date_to=timedelta(days=-4),
    ...     count=3
    ... )
    """

    def __init__(
        self,
        query: str,
        count: int = 10,
        date_from: "OptDateTime" = None,
        date_to: "OptDateTime" = None,
        sort_order: SortOrder = SortOrder.new_to_old,
        extended_params: "ExtendedParams" = None,
    ):
        super().__init__(
            data_type=ContentType.NEWS_HEADLINES_RDP,
            query=query,
            count=count,
            date_from=date_from,
            date_to=date_to,
            sort_order=sort_order,
            extended_params=extended_params,
        )
        self._query = query

    def get_data(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
        on_page_response: Optional[Callable] = None,
    ):
        """
        Returns a response from the API to the library

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data
        on_response : Callable, optional
            Callable object to process retrieved data
        on_page_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        NewsHeadlinesResponse

        Raises
        ------
        AttributeError
            If user didn't set default session.

        Examples
        --------
        >>> from refinitiv.data.content import news
        >>> definition = news.headlines.Definition("Refinitiv",
        >>>                                        date_from="20.03.2021",
        >>>                                        date_to=timedelta(days=-4),
        >>>                                        count=3)
        >>> response = definition.get_data()
        """
        self._kwargs["on_page_response"] = on_page_response
        return super().get_data(session, on_response)

    async def get_data_async(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
        on_page_response: Optional[Callable] = None,
        closure: Optional[str] = None,
    ):
        """
        Returns a response asynchronously from the API to the library

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data
        on_response : Callable, optional
            Callable object to process retrieved data
        on_page_response : Callable, optional
            Callable object to process retrieved data
        closure : str, optional
            Specifies the parameter that will be merged with the request

        Returns
        -------
        NewsHeadlinesResponse

        Raises
        ------
        AttributeError
            If user didn't set default session.

        Examples
        --------
        >>> from refinitiv.data.content import news
        >>> definition = news.headlines.Definition("Refinitiv",
        >>>                                        date_from="20.03.2021",
        >>>                                        date_to=timedelta(days=-4),
        >>>                                        count=3)
        >>> response = await definition.get_data_async()
        """
        self._kwargs["on_page_response"] = on_page_response
        self._kwargs["closure"] = closure
        return await super().get_data_async(session, on_response)

    def __repr__(self):
        return create_repr(
            self,
            middle_path="headlines",
            content=f"{{query='{self._query}'}}",
        )



from refinitiv.data import _configure as configure


def test_watch_enabled():
    # given
    configure.reload()

    # when
    configure._enable_watch()

    # then
    assert configure._observer is not None
    assert configure._observer.should_keep_running() is True


def test_watch_disabled():
    # given
    configure.reload()
    configure._enable_watch()

    # when
    configure._disable_watch()

    # then
    assert configure._observer is not None
    assert configure._observer.should_keep_running() is False


def test_add_listener():
    # given
    configure.reload()
    event_update = configure.ConfigEvent.UPDATE

    # when
    configure.get_config().on(event_update, lambda: None)

    # then
    assert configure.get_config().count(event_update) == 1


def test_remove_listener():
    # given
    configure.reload()
    listener = lambda: None
    event_update = configure.ConfigEvent.UPDATE
    configure.get_config().on(event_update, listener)

    # when
    configure.get_config().remove_listener(event_update, listener)

    # then
    assert configure.get_config().count(event_update) == 0


def test_remove_all_listeners():
    # given
    configure.reload()
    event_update = configure.ConfigEvent.UPDATE
    configure.get_config().on(event_update, lambda: None)

    # when
    configure.get_config().remove_all_listeners()

    # then
    assert len(configure.get_config().listeners(event_update)) == 0


def test_no_watch_than_dispose():
    # given
    configure.reload()
    event_update = configure.ConfigEvent.UPDATE
    configure.get_config().on(event_update, lambda: None)

    # when
    configure._dispose()

    # then
    assert len(configure.get_config().listeners(event_update)) == 0


def test_watch_than_dispose():
    # given
    configure.reload()
    event_update = configure.ConfigEvent.UPDATE
    configure.get_config().on(event_update, lambda: None)
    configure._enable_watch()

    # when
    configure._dispose()

    # then
    assert len(configure.get_config().listeners(event_update)) == 0
    assert configure._observer is not None
    assert configure._observer.should_keep_running() is False



from types import SimpleNamespace

import pytest
from mock import MagicMock

from refinitiv.data import _configure as configure
from refinitiv.data.content._content_provider_layer import ContentProviderLayer
from refinitiv.data._content_type import ContentType
from refinitiv.data._core.session import _rd_default_session_manager
from refinitiv.data._core.session._default_session_manager import (
    Wrapper,
    get_valid_session,
)
from refinitiv.data.delivery._data._data_provider_factory import (
    make_provider,
    get_url,
)
from tests.unit.conftest import StubSession
from tests.unit.content.conftest import StubDefinition


def test_get_url():
    config = configure.get_config()

    url = get_url(ContentType.FORWARD_CURVE, config)
    assert url


def test_get_url_raise():
    config = configure.get_config()
    with pytest.raises(AttributeError):
        get_url(None, config)


def test_make_provider():
    provider = make_provider(ContentType.FORWARD_CURVE)
    assert provider


def test_make_provider_raise():
    with pytest.raises(Exception):
        make_provider(MagicMock())


def test_provider_get_data():
    session = StubSession(is_open=True)
    provider = make_provider(ContentType.FORWARD_CURVE)
    data = provider.get_data(
        session,
        "test_url",
        universe=[MagicMock()],
        __content_type__=ContentType.FORWARD_CURVE,
    )
    assert data


@pytest.mark.asyncio
async def test_provider_get_data_async():
    session = StubSession(is_open=True)
    provider = make_provider(ContentType.FORWARD_CURVE)

    data = await provider.get_data_async(
        session,
        "test_url",
        universe=[MagicMock()],
        universe_class=MagicMock,
        __content_type__=ContentType.FORWARD_CURVE,
    )
    assert data


def test_get_valid_session_has_session_and_has_default():
    session = StubSession()
    default_session = StubSession()
    _rd_default_session_manager.set_session(default_session)

    validated_session = get_valid_session(session)

    assert validated_session == session and validated_session != default_session


def test_get_valid_session_has_session_and_hasnot_default():
    session = StubSession()
    _rd_default_session_manager.set_session(None)

    validated_session = get_valid_session(session)

    assert validated_session == session


def test_get_valid_session_hasnot_session_and_has_default():
    default_session = SimpleNamespace()
    _rd_default_session_manager.set_session(default_session)

    validated_session = get_valid_session(None)

    assert validated_session == default_session

    _rd_default_session_manager.set_session(None)


def test_get_valid_session_hasnot_session_and_hasnot_default():
    _rd_default_session_manager.set_session(None)

    with pytest.raises(AttributeError):
        get_valid_session(None)


def test_content_provider_get_data():
    session = StubSession(is_open=True)
    try:
        provider = ContentProviderLayer(
            ContentType.FORWARD_CURVE, universe=StubDefinition()
        )
        provider.get_data(session, lambda response, inst, session: inst)
    except Exception as e:
        assert False, e


@pytest.mark.asyncio
async def test_content_provider_get_data_async():
    session = StubSession(is_open=True)
    try:
        provider = ContentProviderLayer(
            ContentType.FORWARD_CURVE, universe=StubDefinition()
        )
        await provider.get_data_async(session)
    except Exception as e:
        assert False, e



# coding: utf8

__all__ = ["SurfaceFilters"]

from ..._object_definition import ObjectDefinition

from ._maturity_filter import MaturityFilter
from ._strike_filter_range import StrikeFilterRange
from ._strike_filter import StrikeFilter


class SurfaceFilters(ObjectDefinition):
    def __init__(
        self,
        maturity_filter_range=None,
        strike_range=None,
        strike_range_percent=None,
        atm_tolerance_interval_percent=None,
        ensure_prices_monotonicity=None,
        max_of_median_bid_ask_spread=None,
        max_staleness_days=None,
        use_only_calls=None,
        use_only_puts=None,
        use_weekly_options=None,
    ):
        super().__init__()
        self.maturity_filter_range = maturity_filter_range
        self.strike_range = strike_range
        self.strike_range_percent = strike_range_percent
        self.atm_tolerance_interval_percent = atm_tolerance_interval_percent
        self.ensure_prices_monotonicity = ensure_prices_monotonicity
        self.max_of_median_bid_ask_spread = max_of_median_bid_ask_spread
        self.max_staleness_days = max_staleness_days
        self.use_only_calls = use_only_calls
        self.use_only_puts = use_only_puts
        self.use_weekly_options = use_weekly_options

    @property
    def maturity_filter_range(self):
        """
        Define the MaturityFilterRange
        :return: object MaturityFilter
        """
        return self._get_object_parameter(MaturityFilter, "maturityFilterRange")

    @maturity_filter_range.setter
    def maturity_filter_range(self, value):
        self._set_object_parameter(MaturityFilter, "maturityFilterRange", value)

    @property
    def strike_range(self):
        """
        Define the StrikeFilterRange
        :return: object StrikeFilterRange
        """
        return self._get_object_parameter(StrikeFilterRange, "strikeRange")

    @strike_range.setter
    def strike_range(self, value):
        self._set_object_parameter(StrikeFilterRange, "strikeRange", value)

    @property
    def strike_range_percent(self):
        """
        [DEPRECATED]
        Define the StrikeFilterRange
        :return: object StrikeFilter
        """
        return self._get_object_parameter(StrikeFilter, "strikeRangePercent")

    @strike_range_percent.setter
    def strike_range_percent(self, value):
        self._set_object_parameter(StrikeFilter, "strikeRangePercent", value)

    @property
    def atm_tolerance_interval_percent(self):
        """
        Filter on the ATM tolerance interval percent
        :return: float
        """
        return self._get_parameter("atmToleranceIntervalPercent")

    @atm_tolerance_interval_percent.setter
    def atm_tolerance_interval_percent(self, value):
        self._set_parameter("atmToleranceIntervalPercent", value)

    @property
    def ensure_prices_monotonicity(self):
        """
        Filter on the monotonicity of price options.
        :return: bool
        """
        return self._get_parameter("ensurePricesMonotonicity")

    @ensure_prices_monotonicity.setter
    def ensure_prices_monotonicity(self, value):
        self._set_parameter("ensurePricesMonotonicity", value)

    @property
    def max_of_median_bid_ask_spread(self):
        """
        Spread mutltiplier to filter the options with the same expiry
        :return: float
        """
        return self._get_parameter("maxOfMedianBidAskSpread")

    @max_of_median_bid_ask_spread.setter
    def max_of_median_bid_ask_spread(self, value):
        self._set_parameter("maxOfMedianBidAskSpread", value)

    @property
    def max_staleness_days(self):
        """
        Max Staleness past days to use for building the surface
        :return: int
        """
        return self._get_parameter("maxStalenessDays")

    @max_staleness_days.setter
    def max_staleness_days(self, value):
        self._set_parameter("maxStalenessDays", value)

    @property
    def use_only_calls(self):
        """
        SElect only teh calls to build the surface
        :return: bool
        """
        return self._get_parameter("useOnlyCalls")

    @use_only_calls.setter
    def use_only_calls(self, value):
        self._set_parameter("useOnlyCalls", value)

    @property
    def use_only_puts(self):
        """
        Select only the puts to build the surface
        :return: bool
        """
        return self._get_parameter("useOnlyPuts")

    @use_only_puts.setter
    def use_only_puts(self, value):
        self._set_parameter("useOnlyPuts", value)

    @property
    def use_weekly_options(self):
        """
        Filter on the weekly options.
        :return: bool
        """
        return self._get_parameter("useWeeklyOptions")

    @use_weekly_options.setter
    def use_weekly_options(self, value):
        self._set_parameter("useWeeklyOptions", value)



import json

import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.contracts.conftest import (
    add_call_backs,
)
from tests.integration.content.ipa.contracts.swap.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
)
from tests.integration.content.ipa.contracts.term_deposit.conftest import (
    term_deposit_definition,
    term_deposit_universe,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
)


@allure.suite("Content object - Term Deposit")
@allure.feature("Content object - Term Deposit")
@allure.severity(allure.severity_level.CRITICAL)
class TestTermDeposit:
    @allure.title(
        "Create term deposit definition object with valid params and get data"
    )
    @pytest.mark.caseid("36801954")
    def test_get_term_deposit_analytics_with_valid_params(self, open_session):
        response = term_deposit_definition().get_data()
        check_http_status_is_success_and_df_value_not_empty(response=response)

    @allure.title("Create term deposit definition object without any parameters")
    @pytest.mark.caseid("36801991")
    def test_get_term_deposit_analytics_without_any_params(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.term_deposit.Definition().get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.1011 | Invalid input: instrument definition is invalid. Field 'FixedRatePercent' cannot be empty."
        )

    @allure.title("Create term deposit definition object asynchronously")
    @pytest.mark.caseid("36801992")
    async def test_get_term_deposit_analytics_asynchronously(self, open_session_async):
        response_01, response_02 = await get_async_response_from_definitions(
            rdf.term_deposit.Definition(),
            term_deposit_definition(),
        )

        check_response_status(
            response=response_01,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-DPS.1011",
            expected_error_message="Invalid input: instrument definition is invalid. Field 'FixedRatePercent' cannot be empty.",
        )
        check_http_status_is_success_and_df_value_not_empty(response_02)

    @allure.title("Create term deposit stream with all callbacks")
    @pytest.mark.caseid("36801994")
    @pytest.mark.smoke
    def test_get_term_deposit_analytics(self, open_session):
        events_list = []
        stream = term_deposit_definition().get_stream()
        add_call_backs(stream, events_list)

        stream.open()
        check_stream_state_and_df_from_stream(stream)
        assert "Response received for" in events_list, f"Events list is {events_list}"

    @allure.title("Create a term deposit definition with extended parameters")
    @pytest.mark.caseid("C39506224")
    def test_get_term_deposit_with_extended_params(self, open_session):
        response = rdf.term_deposit.Definition(
            tenor="2Y",
            notional_ccy="USD",
            extended_params={
                "instrumentType": "TermDeposit",
                "instrumentDefinition": {
                    "instrumentTag": "my tag AED_AM1A",
                    "tenor": "5Y",
                    "notionalCcy": "GBP",
                    "fixedRatePercent": 1,
                },
                "pricingParameters": {"valuationDate": "2018-01-10T00:00:00Z"},
            },
            fields=[
                "InstrumentTag",
                "InstrumentDescription",
                "InterestAmountInDealCcy",
                "FixedRate",
                "MarketValueInDealCcy",
                "MarketValueInReportCcy",
                "ErrorMessage",
            ],
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == term_deposit_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.df.InstrumentTag[0] == "my tag AED_AM1A"



import fnmatch
import json
import logging
import os
import sqlite3

import refinitiv.data as rd


def find_all_files_by_pattern(pattern, path):
    results = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                results.append(os.path.join(root, name))
    return results


def delete_file_by_path(path):
    if os.path.exists(path):
        try:
            os.remove(path)
            logging.info("Deleted file: " + path)
        except PermissionError as err:
            logging.error("No permissions to delete the file: " + path)
            logging.error(err)
    else:
        logging.info("File does not exist: " + path)


def check_folder_contains_init_and_delta_files(path):
    files = find_all_files_by_pattern("*jsonl.gz", path)
    assert len(files) >= 2
    assert list_of_strings_contains_pattern(files, "*Init*")
    assert list_of_strings_contains_pattern(files, "*Delta*")


def list_of_strings_contains_pattern(strings_list, pattern):
    for string in strings_list:
        if fnmatch.fnmatch(string, pattern):
            return True
    return False


def parse_file_to_array_of_strings(filepath):
    text_file = open(filepath, "r")
    lines = text_file.readlines()
    text_file.close()
    return lines


def parse_file_to_array_of_dicts(filepath):
    list_of_strings = parse_file_to_array_of_strings(filepath)
    return list(map(lambda string: json.loads(string), list_of_strings))


def check_log_file_has_success_download_lines(path):
    files = find_all_files_by_pattern("*jsonl.gz", path)
    logfile_lines = parse_file_to_array_of_dicts(os.path.join(path, "log.txt"))
    downloaded_files_from_logfile = list(
        filter(lambda line: line["action"] == "DOWNLOAD", logfile_lines)
    )
    downloaded_filenames = list(
        map(lambda dict: dict["details"]["filename"], downloaded_files_from_logfile)
    )
    for file in files:
        assert (
            os.path.basename(file) in downloaded_filenames
        ), f"the line for file download not found in the logfile. {file}. Logfile lines: {logfile_lines}"


def check_log_file_has_success_extract_lines(path):
    files = find_all_files_by_pattern("*jsonl", path)
    logfile_lines = parse_file_to_array_of_dicts(os.path.join(path, "log.txt"))
    extracted_files_from_logfile = list(
        filter(lambda line: line["action"] == "EXTRACT", logfile_lines)
    )
    extracted_filenames = list(
        map(lambda dict: dict["details"]["filename"], extracted_files_from_logfile)
    )
    for file in files:
        assert (
            os.path.basename(file) in extracted_filenames
        ), f"the line for file extract not found in the logfile. {file}. Logfile lines: {logfile_lines}"


def check_downloads_folder(path):
    assert os.path.exists(path), f"The expected path {path} does not exist"
    assert os.path.isdir(path), f"Not a folder: {path}"
    assert len(os.listdir(path)) != 0, f"Folder is empty: {path}"


def check_log_file_has_no_extract_lines(path):
    logfile_lines = parse_file_to_array_of_dicts(os.path.join(path, "log.txt"))
    extracted_files_from_logfile = list(
        filter(lambda line: line["action"] == "EXTRACT", logfile_lines)
    )
    assert (
        len(extracted_files_from_logfile) == 0
    ), f"Found lines about file extraction in the logfile: {extracted_files_from_logfile}"


def check_sqlite_db_data():
    conn = sqlite3.connect("db.sqlite", uri=True)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    table_name = cursor.fetchall()[0][0]
    expected_table_name = rd.get_config().get_param(
        "bulk.esg.standard_scores.db.create-table-queries"
    )[0]["table-name"]
    assert (
        table_name == expected_table_name
    ), f"Invalid table name was created in sqlite db: {table_name}, expected: {expected_table_name}"
    cursor.execute(f"SELECT * FROM {table_name};")
    table_lines = cursor.fetchall()
    assert len(table_lines) > 0, f"Empty data retrieved from sqlite table"


def check_sqlite_db_is_empty():
    conn = sqlite3.connect("db.sqlite", uri=True)
    cursor = conn.cursor()
    configured_table_name = rd.get_config().get_param(
        "bulk.esg.standard_scores.db.create-table-queries"
    )[0]["table-name"]
    cursor.execute(f"SELECT * FROM {configured_table_name};")
    table_lines = cursor.fetchall()
    assert len(table_lines) == 0, f"Sqlite table {configured_table_name} is not empty"


def check_folder_cleaned_up(path):
    JENKINS_SERVICE_FILES_PATTERN = ".nfs"
    list_of_files_in_folder = os.listdir(path)
    filtered_list = list(
        filter(
            lambda x: not x.startswith(JENKINS_SERVICE_FILES_PATTERN),
            list_of_files_in_folder,
        )
    )
    assert len(filtered_list) == 0 or (
        len(filtered_list) == 1 and filtered_list[0] == "log.txt"
    ), f"Folder {path} is not  empty: {list_of_files_in_folder}, list of filtered files is: {filtered_list}"



from datetime import datetime, timedelta

import allure
import pytest

from refinitiv.data.content.ipa.dates_and_calendars import add_periods
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)

EXPECTED_COLUMN_NAMES = [
    "date",
    "tag",
]


first_definition = add_periods.Definition(
    tag="my request 1",
    start_date=timedelta(-180),
    period="4D",
    calendars=["BAR", "KOR", "JAP"],
    currencies=["USD"],
    date_moving_convention="NextBusinessDay",
    end_of_month_convention="Last",
    holiday_outputs=["Date", "Calendars", "Countries", "Names"],
)

second_definition = add_periods.Definition(
    tag="my request 2",
    start_date="2020-04-24",
    period="4D",
    calendars=["BAR", "KOR", "JAP"],
    currencies=["USD"],
    date_moving_convention=add_periods.DateMovingConvention.NEXT_BUSINESS_DAY,
    end_of_month_convention="Last",
    holiday_outputs=[
        add_periods.HolidayOutputs.DATE,
        add_periods.HolidayOutputs.CALENDARS,
        add_periods.HolidayOutputs.NAMES,
    ],
)

invalid_definition = add_periods.Definition(
    tag="my request 2", start_date="2023.04.24", period="inval"
)


@allure.suite("Dates And Calendars")
@allure.feature("Content object - Add Periods")
@allure.severity(allure.severity_level.CRITICAL)
class TestAddPeriods:
    @allure.title("Create a add_periods definition with valid params")
    @pytest.mark.caseid("C45475716")
    @pytest.mark.parametrize(
        "start_date, calendars, currencies, holiday_outputs, end_of_month_convention",
        [
            (
                "2018-12-31",
                ["UKR", "FRA"],
                ["EUR"],
                ["Date", "Names", "Calendars", "Countries"],
                add_periods.EndOfMonthConvention.LAST,
            ),
            (
                timedelta(-180),
                ["UKR", "FRA"],
                None,
                add_periods.HolidayOutputs.NAMES,
                "Same",
            ),
            (timedelta(-180), None, ["USD"], ["Date", "Calendars", "Names"], None),
            (
                datetime(2019, 4, 13),
                None,
                ["USD"],
                None,
                add_periods.EndOfMonthConvention.LAST28,
            ),
        ],
    )
    def test_add_periods_definition_with_valid_params(
        self,
        open_session,
        start_date,
        calendars,
        currencies,
        holiday_outputs,
        end_of_month_convention,
    ):
        response = add_periods.Definition(
            tag="my request",
            period="4D",
            start_date=start_date,
            calendars=calendars,
            currencies=currencies,
            holiday_outputs=holiday_outputs,
            date_moving_convention=add_periods.DateMovingConvention.NEXT_BUSINESS_DAY,
            end_of_month_convention=end_of_month_convention,
        ).get_data()

        assert isinstance(response.data.added_period.holidays, list)
        assert response.data.added_period.tag == "my request"
        assert response.data.added_period.date is not None

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, EXPECTED_COLUMN_NAMES)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create a add_periods definition with valid and invalid definition asynchronously"
    )
    @pytest.mark.caseid("C46335537")
    async def test_add_periods_definition_with_valid_and_invalid_def_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            first_definition, invalid_definition
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
            expected_error_message="Invalid input: at least one currency code or calendar code must be specified.",
        )

    @allure.title(
        "Create a add_periods definitions with valid and invalid definition and getting RDError"
    )
    @pytest.mark.parametrize(
        "expected_error",
        [
            (
                "Error code 400 | Invalid input: at least one currency code or calendar code must be specified."
            )
        ],
    )
    @pytest.mark.caseid("C45475872")
    def test_add_periods_definitions_with_valid_and_invalid_def(
        self, open_session, expected_error
    ):
        with pytest.raises(RDError) as error:
            add_periods.Definitions(
                [first_definition, second_definition, invalid_definition]
            ).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create a add_periods definitions with valid and invalid definition params asynchronously"
    )
    @pytest.mark.caseid("C46335532")
    async def test_add_periods_definitions_with_valid_and_invalid_def_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            add_periods.Definitions([first_definition, second_definition]),
            add_periods.Definitions(invalid_definition),
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
            expected_error_message="Invalid input: at least one currency code or calendar code must be specified.",
        )

    @allure.title("Create add_periods definition with extended params and get data")
    @pytest.mark.parametrize(
        "extended_param",
        [
            {
                "tag": "new tag",
                "startDate": "2022-07-04",
                "holidayOutputs": ["Date"],
            }
        ],
    )
    @pytest.mark.caseid("C45475718")
    def test_add_periods_definition_with_extended_params(
        self, open_session, extended_param
    ):
        response = add_periods.Definition(
            tag="my request",
            period="4D",
            start_date=timedelta(-360),
            calendars=["UKR", "FRA"],
            currencies=["EUR"],
            holiday_outputs=["Date", "Names", "Calendars", "Countries"],
            extended_params=extended_param,
        ).get_data()

        check_extended_params_were_sent_in_request(response, extended_param)
        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

    @allure.title("Create an add_periods definition with closed session and get error")
    @pytest.mark.parametrize(
        "expected_error",
        ["Session is not opened. Can't send any request"],
    )
    @pytest.mark.caseid("C45475868")
    def test_add_periods_definition_with_closed_session(
        self, expected_error, open_platform_session
    ):
        session = open_platform_session
        session.close()
        with pytest.raises(ValueError) as error:
            first_definition.get_data()
        assert str(error.value) == expected_error



import allure
import pytest

from refinitiv.data.content import custom_instruments as ci
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.custom_instruments.conftest import (
    instrument_data,
    new_instrument_data,
    get_holidays,
)
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
)


@allure.suite("Content object - Custom Instrument")
@allure.feature("Formula Custom Instrument")
@allure.severity(allure.severity_level.CRITICAL)
class TestFormulaCustomInstrument:
    @allure.title("Retrieve the data of created Custom Instrument objects")
    @pytest.mark.caseid("C42995813")
    def test_get_data_of_custom_instrument_objects(
        self, open_session_with_rdp_creds_for_ci, create_instrument
    ):
        symbol = create_instrument()
        response = ci.search.Definition().get_data()
        check_response_status(
            response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )
        check_non_empty_response_data(response)
        assert any(
            symbol in universe for universe in response.data.df.get("symbol")
        ), f"There is no created CI symbol on server"

    @allure.title("Create formula Custom Instrument object")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "formula, expected_holidays",
        [
            (
                "VOD.L + LSEG.L",
                [
                    {"date": "2022-09-01", "reason": "Knowledge Day"},
                    {"date": "2022-12-18", "reason": "Hanukkah"},
                ],
            )
        ],
    )
    def test_create_formula_custom_instrument(
        self, open_session_with_rdp_creds_for_ci, symbol, formula, expected_holidays
    ):
        dates_and_calendars_holiday = get_holidays(
            session=open_session_with_rdp_creds_for_ci
        )
        expected_holidays.append(
            {
                "date": dates_and_calendars_holiday.date,
                "reason": dates_and_calendars_holiday.name,
            }
        )

        cust_inst = ci.manage.create_formula(
            **instrument_data,
            symbol=symbol,
            formula=formula,
            holidays=[
                ci.manage.Holiday(date="2022-09-01", name="Knowledge Day"),
                {"date": "2022-12-18", "reason": "Hanukkah"},
                dates_and_calendars_holiday,
            ],
        )

        for param, value in instrument_data.items():
            assert (
                cust_inst.__getattribute__(param) == value
            ), f"Attribute {param} was not created for {symbol}"
        assert (
            symbol in cust_inst.symbol
        ), f"Created symbol is not the same as requested"
        assert cust_inst.holidays == expected_holidays

        ci.manage.delete(symbol)

    @allure.title("Update formula Custom Instrument object with params")
    @pytest.mark.caseid("C42995815")
    @pytest.mark.parametrize(
        "formula, expected_holidays",
        [
            (
                "VOD.L + LSEG.L",
                [
                    {"date": "2022-09-01", "reason": "Knowledge Day"},
                    {"date": "2022-12-18", "reason": "Hanukkah"},
                ],
            )
        ],
    )
    def test_update_formula_custom_instrument_object_with_params(
        self,
        open_session_with_rdp_creds_for_ci,
        create_instrument,
        formula,
        expected_holidays,
    ):
        symbol = create_instrument()
        cust_inst = ci.manage.get(symbol)

        cust_inst.formula = formula
        cust_inst.holidays = expected_holidays

        for param, value in new_instrument_data.items():
            cust_inst.__setattr__(param, value)
        cust_inst.save()

        updated_instrument = ci.manage.get(universe=symbol)

        for param, value in new_instrument_data.items():
            assert (
                updated_instrument.__getattribute__(param) == value
            ), f"Attribute {param} was not updated  for {symbol}"
        assert updated_instrument.formula == formula
        assert updated_instrument.holidays == [
            {"date": "2022-09-01", "reason": "Knowledge Day"},
            {"date": "2022-12-18", "reason": "Hanukkah"},
        ]

    @allure.title(
        "Delete Custom Instrument object and get error for retrieving removed one"
    )
    @pytest.mark.caseid("C42995816")
    def test_delete_custom_instrument_object(
        self, open_session_with_rdp_creds_for_ci, create_instrument
    ):
        symbol = create_instrument()
        ci.manage.delete(symbol)
        with pytest.raises(RDError) as error:
            ci.manage.get(symbol)
        assert (
            str(error.value)
            == f'Error code 404 | CustomInstrument not found: CustomInstrument with "symbol":"{symbol}" not found'
        )

    @allure.title("Create Custom Instrument object with extended params")
    @pytest.mark.parametrize(
        "formula,instrument_params,extended_params",
        [
            (
                "GBP=*3",
                [
                    "description",
                    "exchange_name",
                    "instrument_name",
                    "time_zone",
                    "currency",
                ],
                {
                    "description": "Glory tale about instrument",
                    "exchangeName": "8080",
                    "instrumentName": "Trading St. tool",
                    "timeZone": "LON",
                    "currency": "GBP",
                },
            )
        ],
    )
    @pytest.mark.caseid("C43999085")
    def test_create_custom_instrument_with_extended_params(
        self,
        open_session_with_rdp_creds_for_ci,
        symbol,
        formula,
        instrument_params,
        extended_params,
    ):
        cust_inst = ci.manage.create(
            symbol=symbol, formula=formula, extended_params=extended_params
        )
        for param_key, extended_key in zip(instrument_params, extended_params):
            assert (
                cust_inst.__getattribute__(param_key) == extended_params[extended_key]
            ), f"Attribute {param_key} was not created  for {symbol}"



import allure
import pytest

from refinitiv.data.content.ipa.curves import zc_curves
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.curves.conftest import (
    on_response,
    check_curve_response,
    check_curves,
    zc_curves_01,
    zc_curves_02,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - ZcCurves")
@allure.feature("Content object - ZcCurves")
@allure.severity(allure.severity_level.CRITICAL)
class TestZcCurves:
    @allure.title("Create a zc_curves object with valid definitions and get data")
    @pytest.mark.caseid("39788765")
    @pytest.mark.smoke
    def test_create_zc_curves_definition_with_valid_definitions(self, open_session):
        response = zc_curves.Definitions([zc_curves_01(), zc_curves_02()]).get_data(
            on_response=lambda response, definition, session: on_response(
                response, definition, open_session
            )
        )

        check_curves(response)
        check_curve_response(response, expected_currency="CHF")
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create zc curves object with invalid definition and get data")
    @pytest.mark.caseid("39788766")
    def test_create_invalid_zc_curves(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code QPS-Curves.6 | Invalid input: curveDefinition is missing",
        ):
            zc_curves.Definitions(zc_curves.Definition()).get_data()

    @allure.title("Create a zc curves definition object async")
    @pytest.mark.caseid("39788767")
    async def test_create_zc_curves_definition_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            zc_curves.Definitions([zc_curves_01(), zc_curves_02()]),
            zc_curves.Definitions(zc_curves.Definition()),
        )

        check_curves(valid_response)
        check_curve_response(valid_response, expected_currency="CHF")

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )

    @allure.title(
        "Create a zc curves definition object with valid and invalid zc_crve definition and get data"
    )
    @pytest.mark.caseid("39788768")
    @pytest.mark.smoke
    def test_create_zc_curves_definition_with_valid_and_invalid_forward_curve(
        self, open_session
    ):
        response = zc_curves.Definitions(
            [zc_curves_01(), zc_curves.Definition()]
        ).get_data()

        check_curve_response(response, expected_currency="CHF")
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )
        check_curves(response)



from unittest import mock

from refinitiv.data._content_type import ContentType
from refinitiv.data.content.pricing._stream_facade import Stream, PricingStream
from tests.unit.conftest import StubSession


def test_input_attributes():
    # given
    excepted_attributes = [
        "_session",
        "_always_use_default_session",
        "_universe",
        "_fields",
        "_service",
        "_api",
        "_extended_params",
    ]

    # when
    session = StubSession(is_open=True)
    stream_ = Stream(session)
    attributes = list(stream_.__dict__.keys())

    # then
    assert attributes == excepted_attributes


def test_smoke_stream_into_stream():
    # given
    session = StubSession(is_open=True)

    with mock.patch(
        "refinitiv.data.content.pricing._stream_facade._UniverseStreams.__init__",
        return_value=None,
    ) as mock_stream:
        # when
        s = Stream(session, universe="universe")
        s._stream

        # then
        mock_stream.assert_called_once_with(
            content_type=ContentType.STREAMING_PRICING,
            item_facade_class=PricingStream,
            universe="universe",
            session=session,
            fields=None,
            service=None,
            api=None,
            extended_params=None,
        )



from enum import Enum


class CustomInstrumentTypes(Enum):
    Formula = "formula"
    UDC = "udc"
    Basket = "basket"


class SpreadAdjustmentMethod(Enum):
    CLOSE_TO_CLOSE = "close-to-close"
    OPEN_TO_OPEN = "open-to-open"
    CLOSE_TO_OPEN = "close-to-open"
    CLOSE_TO_OPEN_OLD_GAP = "close-to-open-old-gap"
    CLOSE_TO_OPEN_NEW_GAP = "close-to-open-new-gap"


class VolumeBasedRolloverMethod(Enum):
    VOLUME = "volume"
    OPEN_INTEREST = "openInterest"
    VOLUME_AND_OPEN_INTEREST = "volumeAndOpenInterest"
    VOLUME_OR_OPEN_INTEREST = "volumeOrOpenInterest"


class DayBasedRolloverMethod(Enum):
    DAYS_BEFORE_EXPIRY = "daysBeforeExpiry"
    DAYS_BEFORE_END_OF_MONTH = "daysBeforeEndOfMonth"
    DAYS_AFTER_BEGINNING_OF_MONTH = "daysAfterBeginningOfMonth"



# coding: utf8

from enum import Enum, unique


@unique
class ProjectedIndexCalculationMethod(Enum):
    CONSTANT_COUPON_PAYMENT = "ConstantCouponPayment"
    CONSTANT_INDEX = "ConstantIndex"
    FORWARD_INDEX = "ForwardIndex"



import allure
import pytest

from refinitiv.data.content.esg import full_scores
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "periodenddate",
    "TR.TRESGCScore",
    "TR.TRESGScore",
    "TR.TRESGCControversiesScore",
    "TR.EnvironmentPillarScore",
    "TR.SocialPillarScore",
    "TR.GovernancePillarScore",
    "TR.TRESGResourceUseScore",
    "TR.TRESGEmissionsScore",
    "TR.TRESGInnovationScore",
    "TR.TRESGWorkforceScore",
    "TR.TRESGHumanRightsScore",
    "TR.TRESGCommunityScore",
    "TR.TRESGProductResponsibilityScore",
    "TR.TRESGManagementScore",
    "TR.TRESGShareholdersScore",
    "TR.TRESGCSRStrategyScore",
    "TR.CSRReportingScope",
    "TR.ESGPeriodLastUpdateDate",
]

EXPECTED_TITLES = [
    "Instrument",
    "Period End Date",
    "ESG Combined Score",
    "ESG Score",
    "ESG Controversies Score",
    "Environment Pillar Score",
    "Social Pillar Score",
    "Governance Pillar Score",
    "Resource Use Score",
    "Emissions Score",
    "Innovation Score",
    "Workforce Score",
    "Human Rights Score",
    "Community Score",
    "Product Responsibility Score",
    "Management Score",
    "Shareholders Score",
    "CSR Strategy Score",
    "ESG Reporting Scope",
    "ESG Report Auditor Name",
    "ESG Period Last Update Date",
]


@allure.suite("Content object - ESG Full Scores")
@allure.feature("Content object - ESG Full Scores")
@allure.severity(allure.severity_level.CRITICAL)
class TestESGFullScores:
    @allure.title("Create ESG Full Scores definition object with valid params")
    @pytest.mark.parametrize(
        "universe,start,end,use_field_names_in_headers",
        [("4295904307", 0, -10, True), (["4295904307", "5000002406"], None, -2, False)],
    )
    @pytest.mark.caseid("C40364714")
    @pytest.mark.smoke
    def test_esg_full_scores_definition_object_with_valid_params_and_get_data(
        self, universe, start, end, use_field_names_in_headers, open_platform_session
    ):
        response = full_scores.Definition(
            universe=universe,
            start=start,
            end=end,
            use_field_names_in_headers=use_field_names_in_headers,
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )
        check_non_empty_response_data(response)
        check_dataframe_column_date_for_datetime_type(response)
        if use_field_names_in_headers is True:
            check_response_dataframe_contains_columns_names(
                response, EXPECTED_COLUMN_NAMES
            )
        else:
            check_response_dataframe_contains_columns_names(response, EXPECTED_TITLES)

    @allure.title(
        "Create ESG Full Scores definition object with invalid params and get error"
    )
    @pytest.mark.caseid("C40364716")
    @pytest.mark.parametrize(
        "universe,expected_error",
        [
            (
                [],
                "Error code 400 | Validation error: Missing required parameter 'universe'",
            ),
            (
                ["INVALID"],
                "Error code 412 | Unable to resolve all requested identifiers. Requested items: ['INVALID']",
            ),
        ],
    )
    def test_esg_full_scores_definition_object_with_invalid_params_and_get_error(
        self,
        universe,
        expected_error,
        open_platform_session,
    ):
        with pytest.raises(RDError) as error:
            full_scores.Definition(universe=universe).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create ESG Full Scores definition object with use mix valid and invalid data - asynchronously"
    )
    @pytest.mark.caseid("C40364722")
    @pytest.mark.parametrize(
        "universe,invalid_universe", [(["4295904307", "INVALID"], "INVALID")]
    )
    async def test_esg_full_scores_definition_object_with_use_mix_valid_and_invalid_data(
        self, open_platform_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            full_scores.Definition(universe=universe),
            full_scores.Definition(universe=invalid_universe),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['INVALID']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['INVALID']",
        )



# coding: utf8

from enum import Enum, unique


@unique
class MarketDataAccessDeniedFallback(Enum):
    IGNORE_CONSTITUENTS = "IgnoreConstituents"
    RETURN_ERROR = "ReturnError"
    USE_DELAYED_DATA = "UseDelayedData"



from unittest import mock

import pytest

import refinitiv.data as rd
from refinitiv.data.content import trade_data_service
from tests.unit.conftest import StubSession

mocked_session = mock.MagicMock()

UNIVERSE = [
    {
        "type": "swap",
        "definition": {
            "startDate": "2019-07-28T00:00:00Z",
            "swapType": "Vanilla",
            "tenor": "3Y",
        },
    }
]

FIELDS = ["ContractType", "ContractDate"]
API = "streaming/trading-analytics/redi"


def test_import():
    try:
        from refinitiv.data.content import trade_data_service
    except ImportError:
        assert False


def test_error_import():
    try:
        from refinitiv.data import trade_data_stream
    except ImportError:
        assert True


def test_import_data():
    data = [
        elem for elem in dir(rd.content.trade_data_service) if not elem.startswith("_")
    ]
    assert ["Definition", "Events", "FinalizedOrders", "UniverseTypes"] == data


def test_public_data_definition():
    data = [
        elem
        for elem in dir(rd.content.trade_data_service.Definition)
        if not elem.startswith("_")
    ]
    assert ["get_stream"] == data


def test_definition_creation():
    definition = rd.content.trade_data_service.Definition(
        universe=UNIVERSE, fields=FIELDS
    )

    assert definition._universe == UNIVERSE
    assert definition._fields == FIELDS


def test_get_stream_callbacks():
    list_attr = [
        "open",
        "open_async",
        "close",
        "on_complete",
        "on_add",
        "on_update",
        "on_remove",
        "on_event",
        "on_state",
    ]

    stream = rd.content.trade_data_service.Definition(
        universe=UNIVERSE, fields=FIELDS
    ).get_stream(mocked_session)

    for attr in list_attr:
        assert hasattr(stream, attr), f"{attr} method not exist {type(stream)}"


def test_trade_data_service_raise_error_when_try_open_and_session_is_close():
    # given
    session = StubSession(is_open=True)
    definition = trade_data_service.Definition(
        universe=[],
        finalized_orders=trade_data_service.FinalizedOrders.P1D,
        events=trade_data_service.Events.Full,
    )
    stream = definition.get_stream(session=session)

    # then
    with pytest.raises(AssertionError, match="Session must be open"):
        # when
        session.close()
        stream.open()



from inspect import signature

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession
from tests.unit.conftest import (
    remove_dunder_methods,
    remove_private_attributes,
    has_property_names_in_class,
    get_property_names,
)


def test_ipa_swap_definition_attributes():
    expected_attributes = ["get_data", "get_data_async", "get_stream"]
    testing_attributes = dir(rdf.swap.Definition)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


def test_ipa_swap_attributes():
    expected_attributes = [
        "AdjustInterestToPaymentDate",
        "AmortizationFrequency",
        "AmortizationItem",
        "AmortizationType",
        "BusinessDayConvention",
        "DateRollingConvention",
        "DayCountBasis",
        "Definition",
        "Direction",
        "Frequency",
        "IndexAverageMethod",
        "IndexCompoundingMethod",
        "IndexConvexityAdjustmentIntegrationMethod",
        "IndexConvexityAdjustmentMethod",
        "IndexObservationMethod",
        "IndexResetType",
        "IndexSpreadCompoundingMethod",
        "InputFlow",
        "InterestCalculationConvention",
        "InterestType",
        "LegDefinition",
        "NotionalExchange",
        "PremiumSettlementType",
        "PriceSide",
        "PricingParameters",
        "StubRule",
        "SwaptionType",
        "TenorReferenceDate",
    ]
    testing_attributes = dir(rdf.swap)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, testing_attributes


@pytest.mark.parametrize(
    argnames="input_data",
    ids=["Definition", "LegDefinition", "PricingParameters", "AmortizationItem"],
    argvalues=[
        (
            rdf.swap._swap_definition.SwapInstrumentDefinition,
            {
                "instrument_tag": "instrument_tag",
                "instrument_code": "instrument_code",
                "trade_date": "trade_date",
                "start_date": "start_date",
                "end_date": "end_date",
                "tenor": "tenor",
                "legs": [
                    rdf.swap.LegDefinition(
                        direction=rdf.swap.Direction.RECEIVED,
                        interest_type=rdf.swap.InterestType.STEPPED,
                        notional_ccy="notional_ccy",
                        interest_payment_frequency=rdf.swap.Frequency.BI_MONTHLY,
                        interest_calculation_method=rdf.swap.DayCountBasis.DCB_30_360_US,
                    )
                ],
                "is_non_deliverable": True,
                "settlement_ccy": "settlement_ccy",
                "template": "template",
                "start_tenor": "start_tenor",
            },
        ),
        (
            rdf.swap.LegDefinition,
            {
                "index_spread_compounding_method": rdf.swap.IndexSpreadCompoundingMethod.ISDA_COMPOUNDING,
                "cms_template": "cms_template",
                "floor_strike_percent": 123.123,
                "leg_tag": "leg_tag",
                "direction": rdf.swap.Direction.RECEIVED,
                "interest_type": rdf.swap.InterestType.FIXED,
                "notional_ccy": "notional_ccy",
                "notional_amount": 6.6,
                "fixed_rate_percent": 7.7,
                "index_name": "index_name",
                "index_tenor": "index_tenor",
                "spread_bp": 10.10,
                "interest_payment_frequency": rdf.swap.Frequency.EVERY14_DAYS,
                "interest_calculation_method": rdf.swap.DayCountBasis.DCB_30_360_ISDA,
                "accrued_calculation_method": rdf.swap.DayCountBasis.DCB_30_365_BRAZIL,
                "payment_business_day_convention": rdf.swap.BusinessDayConvention.NEXT_BUSINESS_DAY,
                "payment_roll_convention": rdf.swap.DateRollingConvention.LAST28,
                "index_reset_frequency": rdf.swap.Frequency.BI_MONTHLY,
                "index_reset_type": rdf.swap.IndexResetType.IN_ARREARS,
                "index_fixing_lag": 18,
                "first_regular_payment_date": "first_regular_payment_date",
                "last_regular_payment_date": "last_regular_payment_date",
                "amortization_schedule": [rdf.swap.AmortizationItem()],
                "payment_business_days": "payment_business_days",
                "notional_exchange": rdf.swap.NotionalExchange.BOTH,
                "adjust_interest_to_payment_date": rdf.swap.AdjustInterestToPaymentDate.ADJUSTED,
                "index_compounding_method": rdf.swap.IndexCompoundingMethod.COMPOUNDED,
                "interest_payment_delay": 26,
                "stub_rule": rdf.swap.StubRule.SHORT_FIRST_PRO_RATA.ISSUE,
                "index_fixing_ric": "index_fixing_ric",
                "interest_calculation_convention": rdf.swap.InterestCalculationConvention.BOND_BASIS,
                "index_observation_method": rdf.swap.IndexObservationMethod.LOOKBACK,
                "upfront_amount": 11.11,
                "index_average_method": rdf.swap.IndexAverageMethod.COMPOUNDED_ACTUAL,
                "instrument_tag": "instrument_tag",
            },
        ),
        (
            rdf.swap.PricingParameters,
            {
                "report_ccy": "report_ccy",
                "market_data_date": "market_data_date",
                "index_convexity_adjustment_integration_method": rdf.swap.IndexConvexityAdjustmentIntegrationMethod.RIEMANN_SUM,
                "index_convexity_adjustment_method": rdf.swap.IndexConvexityAdjustmentMethod.BLACK_SCHOLES,
                "discounting_ccy": "discounting_ccy",
                "discounting_tenor": "discounting_tenor",
                "market_value_in_deal_ccy": 4.4,
                "valuation_date": "valuation_date",
                "use_legs_signing": True,
                "tenor_reference_date": rdf.swap.TenorReferenceDate.SPOT_DATE,
                "price_side": rdf.swap.PriceSide.ASK,
            },
        ),
        (
            rdf.swap.AmortizationItem,
            {
                "start_date": "start_date",
                "end_date": "end_date",
                "amortization_frequency": rdf.swap.AmortizationFrequency.ONCE,
                "amortization_type": rdf.swap.AmortizationType.LINEAR,
                "remaining_notional": 1.1,
                "amount": 2.2,
            },
        ),
    ],
)
def test_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    # +1 for (self)
    assert len(s.parameters) == (
        len(args_names) + 1
    ), f"csl={cls}{str(set(get_property_names(cls)) - set(args_names))}"

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "universe": [
            {
                "instrumentDefinition": {
                    "legs": [{"paymentBusinessDays": "payment_business_days"}],
                    "startDate": "start_date",
                },
                "extended_params": "extended_params",
                "instrumentType": "Swap",
                "pricingParameters": {"marketDataDate": "market_data_date"},
            }
        ],
    }
    definition = rdf.swap.Definition(
        start_date="start_date",
        fields=["1", "2"],
        legs=[rdf.swap.LegDefinition(payment_business_days="payment_business_days")],
        pricing_parameters=rdf.swap.PricingParameters(
            market_data_date="market_data_date"
        ),
        extended_params={"extended_params": "extended_params"},
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body



import pytest

from refinitiv.data.content.news._udf_html_parser import HeadlineHTMLParser


@pytest.mark.parametrize(
    ("input_data", "expected_result"),
    [
        (
            '<h1 class="storyHeadline">'
            '\n<span class="headline">Test headline</span> -'
            '\n<span class="source" title="test_title">Test title</span>'
            "\n</h1>",
            {
                "headline": "Test headline",
                "creator": "test_title",
            },
        ),
        (
            '<div class="storyContent" lang="en">'
            '<style type="text/css">'
            ".storyContent * {border-color: inherit !important;outline-color: inherit !important;}</style>"
            '<div class="tr-npp-lead"><p>Text</p><p>for</p></div><div class="tr-npp-body"><p>test</p><p>parser.</p>'
            '<p>Read More</p><p><a href=""data-type="url" class="tr-link" translate="no">AP News Digest 6:30 a.m.'
            '</a></p><p><a href=""data-type="url" class="tr-link" translate="no">...</a></p><p>'
            '<a href=""data-type="url" class="tr-link" translate="no">...</a></p>'
            '</div><p class="line-break"><br/></p><p class="tr-copyright">Copyright</p></div>',
            {
                "text": [
                    "Text",
                    "for",
                    "test",
                    "parser.",
                    "Read More",
                    "AP News Digest 6:30 a.m.",
                    "...",
                    "...",
                    "Copyright",
                ]
            },
        ),
        (
            '<h5 style="direction:ltr"><span data-version-created-date="2021-10-03T11:08:19.096Z" class="releasedDate">03-Oct-2021 11:08:19</span></h5>',
            {"creation_date": "2021-10-03T11:08:19.096Z"},
        ),
    ],
)
def test_udf_html_parser(input_data, expected_result):
    # given
    parser = HeadlineHTMLParser()

    # when
    parser.feed(input_data)
    result = parser.data

    # then
    assert result == expected_result



import abc
from typing import Optional, Dict, Tuple, List, Callable

from pandas import DataFrame

from ._context import Context
from .._containers import FieldsContainer
from .._intervals_consts import NON_INTRA_DAY_INTERVALS
from ..._tools import ohlc
from ..._tools._dataframe import convert_dtypes
from ...content._historical_df_builder import process_historical_raw
from ..._types import StrStrings


class HPContext(Context, abc.ABC):
    @property
    def can_get_data(self) -> bool:
        return bool(self.universe.hp and (not self.fields or self.fields.hp))

    @property
    def is_no_fields(self) -> bool:
        return bool(self.fields.is_no_hp and self.fields.adc == ["TR.RIC"])

    @property
    def can_build_df(self) -> bool:
        return bool(self._hp_data and not (self._adc_data or self._cust_inst_data))

    @property
    def raw(self) -> Optional[Dict]:
        return self._hp_data and self._hp_data.raw

    @property
    def df(self) -> Optional[DataFrame]:
        return self._hp_data and self._hp_data.df

    def build_df(self, *args) -> DataFrame:
        df = convert_dtypes(self.df)
        df.ohlc = ohlc.__get__(df, None)
        return df

    def prepare_to_build(self, interval, *args):
        data, fields = self.prepare_data(interval)
        self._data = data

        if self.is_no_fields:
            if not isinstance(fields, FieldsContainer):
                fields = FieldsContainer(fields)

            self.fields = fields

        self._headers = self.prepare_headers(self.raw)

    def prepare_data(self, interval: str, *args):
        """
        Transform historical pricing data to further merging.

        Transformed data looks like:
        {
            "RIC": [{"Date": "value", "field": "value", "another_field": "value"}],
            "NEXT_RIC": [{"Date": "value", "field": "value", "next_field": "value"}]
        }
        """
        if interval in NON_INTRA_DAY_INTERVALS:

            def parse_date_value(dt):
                return str(dt).split(" ")[0]

        else:

            def parse_date_value(dt):
                return str(dt)

        if isinstance(self.raw, dict):
            return self._parse_dict_to_data_fields(parse_date_value)

        elif isinstance(self.raw, list):
            return self._parse_list_to_data_fields(parse_date_value)

    def _parse_list_to_data_fields(
        self, parse_date_value: Callable
    ) -> Tuple[List[List[dict]], StrStrings]:
        data = [[] for _ in self.universe.hp]
        fields = self.fields.hp

        for index, item in enumerate(self.raw):
            if isinstance(item, list):
                data[index].append({"Type": "pricing", self.date_name: None})

            else:
                if not fields:
                    fields = [header["name"] for header in item["headers"]]

                raw_datas, _, raw_dates = process_historical_raw(item, fields)
                for raw_data, raw_date in zip(raw_datas, raw_dates):
                    item = {
                        "Type": "pricing",
                        self.date_name: parse_date_value(raw_date),
                    }
                    item.update((item for item in zip(fields, raw_data)))
                    data[index].append(item)

        return data, fields

    def _parse_dict_to_data_fields(
        self, parse_date_value: Callable
    ) -> Tuple[List[List[dict]], StrStrings]:
        fields = self.fields.hp
        data = []

        raw_datas, raw_columns, raw_dates = process_historical_raw(self.raw, fields)

        for raw_data, raw_date in zip(raw_datas, raw_dates):
            item = {
                "Type": "pricing",
                self.date_name: parse_date_value(raw_date),
            }

            if fields:
                item.update((item for item in zip(fields, raw_data)))

            else:
                item.update((item for item in zip(raw_columns, raw_data)))

            data.append(item)

        if not fields:
            fields = raw_columns

        lists = [[] for _ in self.universe.hp]
        lists[0] = data
        return lists, fields



cross_universe = {
    "instrumentType": "FxCross",
    "instrumentDefinition": {
        "legs": [
            {
                "tenor": "4Y",
                "dealCcyBuySell": "Buy",
                "contraAmount": 65762500,
                "dealAmount": 1000000,
            }
        ],
        "fxCrossType": "FxNonDeliverableForward",
        "fxCrossCode": "USDINR",
    },
    "pricingParameters": {"valuationDate": "2017-11-15T00:00:00Z"},
}



# coding: utf8

from enum import Enum, unique


@unique
class Rounding(Enum):
    DEFAULT = "Default"
    EIGHT = "Eight"
    FIVE = "Five"
    FOUR = "Four"
    ONE = "One"
    SEVEN = "Seven"
    SIX = "Six"
    THREE = "Three"
    TWO = "Two"
    UNROUNDED = "Unrounded"
    ZERO = "Zero"



# coding: utf8

from enum import Enum, unique


@unique
class Frequency(Enum):
    EVERYDAY = "Everyday"
    BI_MONTHLY = "BiMonthly"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    SEMI_ANNUAL = "SemiAnnual"
    ANNUAL = "Annual"
    EVERY7_DAYS = "Every7Days"
    EVERY14_DAYS = "Every14Days"
    EVERY28_DAYS = "Every28Days"
    EVERY30_DAYS = "Every30Days"
    EVERY91_DAYS = "Every91Days"
    EVERY182_DAYS = "Every182Days"
    EVERY364_DAYS = "Every364Days"
    EVERY365_DAYS = "Every365Days"
    EVERY90_DAYS = "Every90Days"
    EVERY92_DAYS = "Every92Days"
    EVERY93_DAYS = "Every93Days"
    EVERY180_DAYS = "Every180Days"
    EVERY183_DAYS = "Every183Days"
    EVERY184_DAYS = "Every184Days"
    EVERY4_MONTHS = "Every4Months"
    EVERY_WORKING_DAY = "EveryWorkingDay"
    R2 = "R2"
    R4 = "R4"
    SCHEDULED = "Scheduled"
    ZERO = "Zero"



from refinitiv.data._errors import RDError
from refinitiv.data.content._content_data import Data
from refinitiv.data.content.ipa._enums import HolidayOutputs
from refinitiv.data.content.ipa.dates_and_calendars import is_working_day
from refinitiv.data.content.ipa.dates_and_calendars.is_working_day._is_working_day_data_provider import (
    IsWorkingDays,
    IsWorkingDay,
)
from tests.unit.conftest import StubSession, StubResponse, remove_private_attributes


def test_is_working_days_class():
    # fmt: off
    raw_response = [
    {"isWorkingDay": False, "isWeekEnd": True, "tag": "my request",
        "holidays": [{"date": "2022-01-01",
                "names": [
                    {"name": "New Year's Day", "calendars": ["BAR", "KOR", "JAP", "USA"],
                     "countries": ["BLR", "KOR", "JPN", "USA"],}],}
        ],},
    {"isWorkingDay": False, "isWeekEnd": True, "tag": "my request",
        "holidays": [{"date": "2021-12-25",
                "names": [
                    {"name": "Christmas Day", "calendars": ["BAR", "EMU"],
                     "countries": ["BLR", ""],}],}
        ],},
    ]
    # fmt: on
    session = StubSession(is_open=True, response=StubResponse(raw_response))
    first_definition = is_working_day.Definition(
        tag="my request",
        date="2022-01-01",
        calendars=["BAR", "KOR", "JAP"],
        currencies=["USD"],
        holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
    )
    second_definition = is_working_day.Definition(
        tag="my request",
        date="2021-12-25",
        calendars=["BAR"],
        currencies=["EUR"],
        holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
    )
    definition = is_working_day.Definitions([first_definition, second_definition])

    response = definition.get_data(session)
    data = response.data

    assert data.raw == raw_response
    assert data.df.empty is False
    assert data.days
    assert isinstance(data, IsWorkingDays)

    assert remove_private_attributes(dir(data)) == ["days", "df", "raw"]


def test_is_working_day_class():
    # fmt: off
    raw_response = [
    {"isWorkingDay": False, "isWeekEnd": True, "tag": "my request",
        "holidays": [{"date": "2021-12-25",
                "names": [
                    {"name": "Christmas Day", "calendars": ["BAR", "EMU"],
                     "countries": ["BLR", ""],}],}
        ],},
    ]
    # fmt: on
    session = StubSession(is_open=True, response=StubResponse(raw_response))
    definition = is_working_day.Definition(
        tag="my request",
        date="2021-12-25",
        calendars=["BAR"],
        currencies=["EUR"],
        holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
    )
    response = definition.get_data(session)

    data = response.data
    assert data.raw == raw_response
    assert data.df.empty is False
    assert data.day
    assert isinstance(data, IsWorkingDay)

    assert remove_private_attributes(dir(data)) == ["day", "df", "raw"]


def test_data_class_when_get_data():
    # fmt: off
    raw_response = [
        {
            "error": {
                "status": "Error",
                "message": "Business Calendar: calendar not found.",
                "code": "IPA_library.ErrorBusinessCalendar_CalendarNotFound",
            },
            "tag": "First Jan for UAH currency",
        }
    ]
    # fmt: on
    session = StubSession(is_open=True, response=StubResponse(raw_response))
    definition = is_working_day.Definition(
        tag="my request",
        date="2021-12-25",
        calendars=["BAR"],
        currencies=["EUR"],
        holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
    )

    try:
        response = definition.get_data(session)
    except RDError as e:
        response = e.response
        assert e
    else:
        assert False

    data = response.data
    assert data.raw == raw_response
    assert data.df.empty is True
    assert isinstance(data, Data)

    assert remove_private_attributes(dir(data)) == ["df", "raw"]


async def test_data_class_when_get_data_async():
    # fmt: off
    raw_response = [
        {
            "error": {
                "status": "Error",
                "message": "Business Calendar: calendar not found.",
                "code": "IPA_library.ErrorBusinessCalendar_CalendarNotFound",
            },
            "tag": "First Jan for UAH currency",
        }
    ]
    # fmt: on
    session = StubSession(is_open=True, response=StubResponse(raw_response))
    definition = is_working_day.Definition(
        tag="my request",
        date="2021-12-25",
        calendars=["BAR"],
        currencies=["EUR"],
        holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
    )
    response = await definition.get_data_async(session)

    data = response.data
    assert data.raw == raw_response
    assert data.df.empty is True
    assert isinstance(data, Data)

    assert remove_private_attributes(dir(data)) == ["df", "raw"]



import pytest

import refinitiv.data as rd
import refinitiv.data.content.esg.bulk as bulk


@pytest.fixture(scope="function", autouse=False)
def prepare_file_manager_and_clean_old_files():
    file_manager = bulk.PackageManager("esg.standard_scores")
    file_manager.cleanup_files()
    yield file_manager


@pytest.fixture(scope="function", autouse=False)
def set_auto_extract_to_false():
    rd.get_config().set_param(
            param=f"bulk.esg.standard_scores.package.download.auto-extract",
            value=False,
            auto_create=True,
        )



import allure
import pytest

from refinitiv.data.content import news
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.news.conftest import (
    check_every_story_title_contains_keyword,
    check_count_in_response_equal_to,
    check_data_in_response_sorted_by_order,
    check_dates_range_in_response,
    check_extended_params_were_sent_in_news_request, check_news_headlines_date_for_datetime_type,
)
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
    get_async_response_from_definition,
    check_index_column_contains_dates,
)


@allure.suite("Content object - News Headlines")
@allure.feature("Content object - News Headlines")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("news")
class TestNewsHeadlines:
    @allure.title("Create News Headlines object with headline query and call get_data")
    @pytest.mark.caseid("34114877")
    @pytest.mark.parametrize("query", ["Google"])
    @pytest.mark.smoke
    def test_create_news_headlines_object_and_get_data(self, open_session, query):
        response = news.headlines.Definition(query).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_every_story_title_contains_keyword(response, query)
        check_news_headlines_date_for_datetime_type(response)

    @allure.title("Create News Headlines object using closed session")
    @pytest.mark.caseid("34114878")
    @pytest.mark.parametrize("query", ["Microsoft"])
    def test_create_content_object_search_when_session_is_not_opened(
        self, query, open_session
    ):
        session = open_session
        session.close()
        with pytest.raises(ValueError) as error:
            news.headlines.Definition(query).get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title("Create News headlines object with specified count param")
    @pytest.mark.caseid("34114877")
    @pytest.mark.parametrize("query,count,expected_raw_count", [("SpaceX", 150, 200)])
    def test_create_news_headlines_object_with_specified_count_param(
        self, open_session, query, count, expected_raw_count
    ):
        session = open_session
        session_type = session.type.name
        response = news.headlines.Definition(query, count=count).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_count_in_response_equal_to(
            count, response, session_type, expected_raw_count
        )

    @allure.title(
        "News headlines object with specified sort_order param new-to-old/old-to-new "
    )
    @pytest.mark.caseid("34114880")
    @pytest.mark.parametrize(
        "query,sort_order",
        [
            ("Tesla", news.headlines.SortOrder.new_to_old),
            ("Tesla", news.headlines.SortOrder.old_to_new),
        ],
    )
    def test_create_news_headlines_object_with_specified_sort_order_param(
        self, open_platform_session, query, sort_order
    ):
        response = news.headlines.Definition(query, sort_order=sort_order).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_data_in_response_sorted_by_order(response, sort_order)

    @allure.title(
        "News headlines object with specified date_from and date_to parameters"
    )
    @pytest.mark.caseid("34114882")
    @pytest.mark.parametrize(
        "query,date_from,date_to,count",
        [("Tesla", "2021-05-15T13:12:33.307Z", "2021-11-29T15:15:10.006Z", 115)],
    )
    def test_create_news_headlines_object_with_specified_date_from_and_date_to(
        self,
        open_session,
        query,
        date_from,
        date_to,
        count,
    ):
        response = news.headlines.Definition(
            query, date_from=date_from, date_to=date_to, count=count
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_dates_range_in_response(response, date_from, date_to)
        check_index_column_contains_dates(response)
        assert (
            response.data.df.index.size == count
        ), f"Actual DF size is {response.data.df.index.size}"

    @allure.title("News headlines object with specified extended_params")
    @pytest.mark.caseid("34114881")
    @pytest.mark.parametrize(
        "query,count,overridden_query,overridden_count",
        [("Microsoft", 5, "Google", "20")],
    )
    def test_create_news_headlines_object_with_specified_extended_params(
        self,
        open_session,
        query,
        count,
        overridden_query,
        overridden_count,
    ):
        extended_params = {"query": overridden_query, "number": overridden_count}
        response = news.headlines.Definition(
            query, count=count, extended_params=extended_params
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_extended_params_were_sent_in_news_request(response, extended_params)
        check_every_story_title_contains_keyword(response, overridden_query)

    @allure.title(
        "Create News Headlines object with headline query and call get_data_async "
    )
    @pytest.mark.caseid("35150253")
    @pytest.mark.parametrize("query", ["refinitiv"])
    async def test_create_news_headlines_object_and_get_data_async(
        self, open_session_async, query
    ):
        response = await get_async_response_from_definition(
            news.headlines.Definition(query)
        )
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_every_story_title_contains_keyword(response, query)

    @allure.title("Create News Headlines object using closed session (asynchronously)")
    @pytest.mark.caseid("35150254")
    @pytest.mark.parametrize("query", ["Microsoft"])
    async def test_create_news_headlines_object_when_session_is_not_opened_async(
        self, query, open_session_async
    ):
        session = open_session_async
        await session.close_async()
        news_definition = news.headlines.Definition(query)
        with pytest.raises(ValueError) as error:
            await get_async_response_from_definition(news_definition)
        assert str(error.value) == "Session is not opened. Can't send any request"



from typing import TYPE_CHECKING

from ._stream_facade import Stream
from .._content_data import Data
from .._content_provider_layer import ContentUsageLoggerMixin
from ..._content_type import ContentType
from ..._core.session import Session
from ..._tools import create_repr, try_copy_to_list
from ..._tools._common import universe_arg_parser, fields_arg_parser
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ..._types import OptStr, ExtendedParams, StrStrings, OptStrStrs


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]], DataProviderLayer[BaseResponse[Data]]
):
    """
    This class defines parameters for requesting events from pricing

    Parameters
    ----------
    universe : str or list of str
        The single/multiple instrument/s name (e.g. "EUR=" or ["EUR=", "CAD=", "UAH="]).
    fields : str or list of str, optional
        Specifies the specific fields to be delivered when messages arrive
    service : str, optional
        Name of the streaming service publishing the instruments
    api: str, optional
        Specifies the data source. It can be updated/added using config file
    extended_params : dict, optional
        Other parameters can be provided if necessary

    Examples
    --------
    >>> from refinitiv.data.content import pricing
    >>> definition = pricing.Definition("EUR=")
    >>> response = definition.get_data()

    """

    _USAGE_CLS_NAME = "Pricing.PricingDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        fields: "OptStrStrs" = None,
        service: "OptStr" = None,
        api: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        extended_params = extended_params or {}
        universe = extended_params.pop("universe", universe)
        universe = try_copy_to_list(universe)
        universe = universe_arg_parser.get_list(universe)
        fields = extended_params.pop("fields", fields)
        fields = try_copy_to_list(fields)
        fields = fields_arg_parser.get_unique(fields or [])
        super().__init__(
            data_type=ContentType.PRICING,
            universe=universe,
            fields=fields,
            extended_params=extended_params,
        )
        self._universe = universe
        self._fields = fields
        self._service = service
        self._api = api
        self._extended_params = extended_params

    def __repr__(self) -> str:
        return create_repr(
            self,
            content=f"{{name={self._universe}}}",
        )

    def get_stream(self, session: Session = None) -> Stream:
        """
        Summary line of this func create a pricing.Stream object for the defined data

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data

        Returns
        -------
        pricing.Stream

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("IBM")
        >>> stream = definition.get_stream()
        >>> stream.open()
        """
        return Stream(
            universe=self._universe,
            session=session,
            fields=self._fields,
            service=self._service,
            api=self._api,
            extended_params=self._extended_params,
        )



from typing import Optional, TYPE_CHECKING

from ......_tools import try_copy_to_list
from ..._bond_curves._enums import (
    BusinessSector,
    CurveSubType,
    EconomicSector,
    Industry,
    IndustryGroup,
    IssuerType,
    MainConstituentAssetClass,
    Rating,
    RatingScaleSource,
    ReferenceEntityType,
    Seniority,
)
from ...._object_definition import ObjectDefinition

if TYPE_CHECKING:
    from ......_types import OptStrings, OptStr


class CreditCurveDefinition(ObjectDefinition):
    """
    Generates the credit curves for the definition provided

    Parameters
    ----------
    business_sector : BusinessSector, optional
        Trbc business sector of the economic sector.
    credit_curve_type_fallback_logic : string, optional
        Credit curve types list used to define the fallback logic in order to find the
        best credit curve. curves types are ordered by priority.
    curve_sub_type : CurveSubType, optional

    curve_tenors : string, optional
        User defined maturities to compute.
    economic_sector : EconomicSector, optional
        Trbc economic sector of the issuer.  available values are: basicmaterials,
        consumer cyclicals, consumer non-cyclicals, energy, financials, healthcare,
        industrials, technology, utilities
    industry : Industry, optional
        Trbc industry of the industry group.
    industry_group : IndustryGroup, optional
        Trbc industry group of the business sector.
    issuer_type : IssuerType, optional
        Type of the issuer. available values are: agency, corporate, munis,
        nonfinancials, sovereign, supranational
    main_constituent_asset_class : MainConstituentAssetClass, optional

    rating : Rating, optional
        Rating of the issuer. the rating can be defined by using any of: "refinitiv",
        "s&p", "moody's", "fitch", "dbrs" convention
    rating_scale_source : RatingScaleSource, optional

    reference_entity_type : ReferenceEntityType, optional
        Type of the reference entity (mandatory if referenceentity is defined).
        avialable values are:   - chainric   - bondisin   - bondric   - organisationid
        - ticker
    seniority : Seniority, optional
        Type of seniority. available values are: senior preferred, subordinate
        unsecured, senior non-preferred
    country : str, optional
        Country code of the issuer defined with alpha-2 code iso 3166 country code
        convention
    currency : str, optional
        Bond curve currency code
    id : str, optional

    name : str, optional

    reference_entity : str, optional
        Code to define the reference entity.
    source : str, optional
        Source or contributor code default value is "refinitiv"
    """

    def __init__(
        self,
        business_sector: Optional[BusinessSector] = None,
        credit_curve_type_fallback_logic: "OptStrings" = None,
        curve_sub_type: Optional[CurveSubType] = None,
        curve_tenors: "OptStrings" = None,
        economic_sector: Optional[EconomicSector] = None,
        industry: Optional[Industry] = None,
        industry_group: Optional[IndustryGroup] = None,
        issuer_type: Optional[IssuerType] = None,
        main_constituent_asset_class: Optional[MainConstituentAssetClass] = None,
        rating: Optional[Rating] = None,
        rating_scale_source: Optional[RatingScaleSource] = None,
        reference_entity_type: Optional[ReferenceEntityType] = None,
        seniority: Optional[Seniority] = None,
        country: "OptStr" = None,
        currency: "OptStr" = None,
        id: "OptStr" = None,
        name: "OptStr" = None,
        reference_entity: "OptStr" = None,
        source: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.business_sector = business_sector
        self.credit_curve_type_fallback_logic = credit_curve_type_fallback_logic
        self.curve_sub_type = curve_sub_type
        self.curve_tenors = try_copy_to_list(curve_tenors)
        self.economic_sector = economic_sector
        self.industry = industry
        self.industry_group = industry_group
        self.issuer_type = issuer_type
        self.main_constituent_asset_class = main_constituent_asset_class
        self.rating = rating
        self.rating_scale_source = rating_scale_source
        self.reference_entity_type = reference_entity_type
        self.seniority = seniority
        self.country = country
        self.currency = currency
        self.id = id
        self.name = name
        self.reference_entity = reference_entity
        self.source = source

    @property
    def business_sector(self):
        """
        Trbc business sector of the economic sector.
        :return: enum BusinessSector
        """
        return self._get_enum_parameter(BusinessSector, "businessSector")

    @business_sector.setter
    def business_sector(self, value):
        self._set_enum_parameter(BusinessSector, "businessSector", value)

    @property
    def credit_curve_type_fallback_logic(self):
        """
        Credit curve types list used to define the fallback logic in order to find the
        best credit curve. curves types are ordered by priority.
        :return: list string
        """
        return self._get_list_parameter(str, "creditCurveTypeFallbackLogic")

    @credit_curve_type_fallback_logic.setter
    def credit_curve_type_fallback_logic(self, value):
        self._set_list_parameter(str, "creditCurveTypeFallbackLogic", value)

    @property
    def curve_sub_type(self):
        """
        :return: enum CurveSubType
        """
        return self._get_enum_parameter(CurveSubType, "curveSubType")

    @curve_sub_type.setter
    def curve_sub_type(self, value):
        self._set_enum_parameter(CurveSubType, "curveSubType", value)

    @property
    def curve_tenors(self):
        """
        User defined maturities to compute.
        :return: list string
        """
        return self._get_list_parameter(str, "curveTenors")

    @curve_tenors.setter
    def curve_tenors(self, value):
        self._set_list_parameter(str, "curveTenors", value)

    @property
    def economic_sector(self):
        """
        Trbc economic sector of the issuer.  available values are: basicmaterials,
        consumer cyclicals, consumer non-cyclicals, energy, financials, healthcare,
        industrials, technology, utilities
        :return: enum EconomicSector
        """
        return self._get_enum_parameter(EconomicSector, "economicSector")

    @economic_sector.setter
    def economic_sector(self, value):
        self._set_enum_parameter(EconomicSector, "economicSector", value)

    @property
    def industry(self):
        """
        Trbc industry of the industry group.
        :return: enum Industry
        """
        return self._get_enum_parameter(Industry, "industry")

    @industry.setter
    def industry(self, value):
        self._set_enum_parameter(Industry, "industry", value)

    @property
    def industry_group(self):
        """
        Trbc industry group of the business sector.
        :return: enum IndustryGroup
        """
        return self._get_enum_parameter(IndustryGroup, "industryGroup")

    @industry_group.setter
    def industry_group(self, value):
        self._set_enum_parameter(IndustryGroup, "industryGroup", value)

    @property
    def issuer_type(self):
        """
        Type of the issuer. available values are: agency, corporate, munis,
        nonfinancials, sovereign, supranational
        :return: enum IssuerType
        """
        return self._get_enum_parameter(IssuerType, "issuerType")

    @issuer_type.setter
    def issuer_type(self, value):
        self._set_enum_parameter(IssuerType, "issuerType", value)

    @property
    def main_constituent_asset_class(self):
        """
        :return: enum MainConstituentAssetClass
        """
        return self._get_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass"
        )

    @main_constituent_asset_class.setter
    def main_constituent_asset_class(self, value):
        self._set_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass", value
        )

    @property
    def rating(self):
        """
        Rating of the issuer. the rating can be defined by using any of: "refinitiv",
        "s&p", "moody's", "fitch", "dbrs" convention
        :return: enum Rating
        """
        return self._get_enum_parameter(Rating, "rating")

    @rating.setter
    def rating(self, value):
        self._set_enum_parameter(Rating, "rating", value)

    @property
    def rating_scale_source(self):
        """
        :return: enum RatingScaleSource
        """
        return self._get_enum_parameter(RatingScaleSource, "ratingScaleSource")

    @rating_scale_source.setter
    def rating_scale_source(self, value):
        self._set_enum_parameter(RatingScaleSource, "ratingScaleSource", value)

    @property
    def reference_entity_type(self):
        """
        Type of the reference entity (mandatory if referenceentity is defined).
        avialable values are:   - chainric   - bondisin   - bondric   - organisationid
        - ticker
        :return: enum ReferenceEntityType
        """
        return self._get_enum_parameter(ReferenceEntityType, "referenceEntityType")

    @reference_entity_type.setter
    def reference_entity_type(self, value):
        self._set_enum_parameter(ReferenceEntityType, "referenceEntityType", value)

    @property
    def seniority(self):
        """
        Type of seniority. available values are: senior preferred, subordinate
        unsecured, senior non-preferred
        :return: enum Seniority
        """
        return self._get_enum_parameter(Seniority, "seniority")

    @seniority.setter
    def seniority(self, value):
        self._set_enum_parameter(Seniority, "seniority", value)

    @property
    def country(self):
        """
        Country code of the issuer defined with alpha-2 code iso 3166 country code
        convention
        :return: str
        """
        return self._get_parameter("country")

    @country.setter
    def country(self, value):
        self._set_parameter("country", value)

    @property
    def currency(self):
        """
        Bond curve currency code
        :return: str
        """
        return self._get_parameter("currency")

    @currency.setter
    def currency(self, value):
        self._set_parameter("currency", value)

    @property
    def id(self):
        """
        :return: str
        """
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)

    @property
    def name(self):
        """
        :return: str
        """
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)

    @property
    def reference_entity(self):
        """
        Code to define the reference entity.
        :return: str
        """
        return self._get_parameter("referenceEntity")

    @reference_entity.setter
    def reference_entity(self, value):
        self._set_parameter("referenceEntity", value)

    @property
    def source(self):
        """
        Source or contributor code default value is "refinitiv"
        :return: str
        """
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)



from typing import Dict

from .._tools import cached_property
from .._content_type import ContentType
from ..content._universe_stream import _UniverseStream
from ..content._universe_streams import _UniverseStreams
from ..content.custom_instruments._custom_instruments_data_provider import (
    is_instrument_id,
    get_user_id,
    symbol_with_user_id,
)
from ..content.custom_instruments._stream_facade import uuid_from_symbol
from ..content.pricing._stream_facade import Stream as _Stream, PricingStream
from ..delivery._data._data_provider import DataProviderLayer
from ..delivery._data._endpoint_data import RequestMethod
from ..delivery._stream import StreamStateEvent


class MixedStreams(_UniverseStreams):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, content_type=ContentType.NONE)
        self._uuid = None

    def _get_symbol(self, universe):
        if is_instrument_id.match(universe):
            data_provider_layer = DataProviderLayer(
                data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
                universe=self.universe,
            )
            instrument_response = data_provider_layer.get_data(
                self._session, method=RequestMethod.GET
            )
            symbol = instrument_response.data.raw.get("symbol")
            if not self._uuid:
                self._uuid = uuid_from_symbol(symbol)
        else:
            if not symbol_with_user_id.match(universe):
                if not self._uuid:
                    self._uuid = get_user_id(self._session)
                symbol = f"{universe}.{self._uuid}"
            else:
                symbol = universe
        return symbol

    def _get_pricing_stream(self, name):
        stream = _UniverseStream(
            content_type=ContentType.STREAMING_PRICING,
            session=self._session,
            name=name,
            fields=self.fields,
            service=self._service,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
            extended_params=self._extended_params,
            parent_id=self._id,
        )
        return stream

    def _get_custom_instruments_stream(self, name):
        stream = _UniverseStream(
            content_type=ContentType.STREAMING_CUSTOM_INSTRUMENTS,
            session=self._session,
            name=self._get_symbol(name),
            fields=self.fields,
            service=self._service,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
            extended_params=self._extended_params,
            parent_id=self._id,
        )
        return stream

    def _create_stream_by_name(self, name):
        if name.startswith("S)"):
            stream = self._get_custom_instruments_stream(name)
        else:
            stream = self._get_pricing_stream(name)
        return stream

    @cached_property
    def _stream_by_name(self) -> Dict[str, _UniverseStream]:
        retval = {}
        for name in self.universe:
            stream = self._create_stream_by_name(name)
            stream.on(StreamStateEvent.CLOSED, self._on_stream_close)
            retval[name] = stream
        return retval

    def add_instruments(self, *instruments):
        instruments = [
            self._get_symbol(name) if name.startswith("S)") else name for name in instruments
        ]
        super().add_instruments(*instruments)

    def remove_instruments(self, *args):
        args = [
            self._get_symbol(name) if name.startswith("S)") else name for name in args
        ]
        super().remove_instruments(*args)


class Stream(_Stream):
    @cached_property
    def _stream(self) -> MixedStreams:
        return MixedStreams(
            item_facade_class=PricingStream,
            universe=self._universe,
            session=self._session,
            fields=self._fields,
            service=self._service,
            extended_params=self._extended_params,
        )



from ...._curves._cross_currency_curves._enums import (
    InterpolationMode,
    MainConstituentAssetClass,
    RiskType,
)
from ...._curves._cross_currency_curves._definitions._enums import (
    QuotationMode,
    StandardTurnPeriod,
)



from ...._tools import add_periods_datetime_adapter
from .._object_definition import ObjectDefinition


class StartEndDateBase(ObjectDefinition):
    def __init__(
        self,
        start_date,
        end_date,
    ):
        super().__init__()
        self.start_date = start_date
        self.end_date = end_date

    @property
    def start_date(self):
        """
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        if value is not None:
            value = add_periods_datetime_adapter.get_str(value)
        self._set_parameter("startDate", value)

    @property
    def end_date(self):
        """
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        if value is not None:
            value = add_periods_datetime_adapter.get_str(value)
        self._set_parameter("endDate", value)


class StartDateBase(ObjectDefinition):
    def __init__(
        self,
        start_date,
    ):
        super().__init__()
        self.start_date = start_date

    @property
    def start_date(self):
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        if value is not None:
            value = add_periods_datetime_adapter.get_str(value)
        self._set_parameter("startDate", value)


class DateBase(ObjectDefinition):
    def __init__(
        self,
        date,
    ):
        super().__init__()
        self.date = date

    @property
    def date(self):
        return self._get_parameter("date")

    @date.setter
    def date(self, value):
        if value is not None:
            value = add_periods_datetime_adapter.get_str(value)
        self._set_parameter("date", value)



import numpy as np
import pytest
from pandas.core.dtypes.common import is_datetime64_any_dtype

import refinitiv.data as rd
from refinitiv.data._content_type import ContentType
from refinitiv.data.content import esg
from refinitiv.data.content.esg._base_definition import get_package_name
from refinitiv.data.content.esg._esg_data_provider import ESGRequestFactory
from tests.unit.conftest import StubResponse, StubSession


def test_esg_request_factory():
    # given
    expected_value = [("universe", "universe"), ("start", 0), ("end", 0)]
    factory = ESGRequestFactory()

    # when
    testing_value = factory.get_query_parameters(universe="universe", start=0, end=0)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "esg_definition, expected_repr",
    [
        (
            rd.content.esg.basic_overview.Definition(universe="IBM111.N"),
            "<refinitiv.data.content.esg.basic_overview.Definition object at {0} {{universe='IBM111.N'}}>",
        ),
        (
            rd.content.esg.full_measures.Definition(universe="BNPP.PA"),
            "<refinitiv.data.content.esg.full_measures.Definition object at {0} {{universe='BNPP.PA', start='None', end='None', closure='None'}}>",
        ),
        (
            rd.content.esg.full_scores.Definition(universe="5000002406"),
            "<refinitiv.data.content.esg.full_scores.Definition object at {0} {{universe='5000002406', start='None', end='None', closure='None'}}>",
        ),
        (
            rd.content.esg.standard_measures.Definition(universe="BNPP.PA"),
            "<refinitiv.data.content.esg.standard_measures.Definition object at {0} {{universe='BNPP.PA', start='None', end='None', closure='None'}}>",
        ),
        (
            rd.content.esg.standard_scores.Definition(universe="5000002406"),
            "<refinitiv.data.content.esg.standard_scores.Definition object at {0} {{universe='5000002406', start='None', end='None', closure='None'}}>",
        ),
        (
            rd.content.esg.universe.Definition(),
            "<refinitiv.data.content.esg.universe.Definition object at {0} {{closure='None'}}>",
        ),
    ],
)
def test_esg_content_repr(esg_definition, expected_repr):
    # given
    obj_id = hex(id(esg_definition))

    # when
    s = repr(esg_definition)

    # then
    assert s == expected_repr.format(obj_id)


def test_universe_get_data_when_session_is_not_open_will_raise_error():
    definition = rd.content.esg.universe.Definition()

    with pytest.raises(AttributeError):
        rd.session.set_default(None)
        definition.get_data()


def test_get_package_name_return_value():
    # given
    input_value = ContentType.ESG_STANDARD_SCORES

    # when
    package_name = get_package_name(input_value)

    # then
    assert package_name


def test_get_package_name_raise_error_if_not_exists():
    # given
    input_value = "invalid value"

    # then
    with pytest.raises(Exception):
        # when
        get_package_name(input_value)


def test_header_period_end_date_has_type_date_and_we_convert_it_to_datetime64():
    # fmt: off
    response = StubResponse({
            "links": {"count": 1}, "variability": "variable", "universe": [{
            "Instrument": "IBM.N",
            "Company Common Name": "International Business Machines Corp",
            "Organization PermID": "4295904307",
            "Reporting Currency": "USD"
        }],
            "data": [
                ["IBM.N", "2021-12-31", 100, "2023-01-12T00:00:00", 780567, 31.7, 84]],
            "messages": {
                "codes": [[-1, -1, -1, -1, -1, -1, -1]],
                "descriptions": [{"code": -1, "description": "ok"}]
            }, "headers": [{
            "name": "instrument", "title": "Instrument", "type": "string",
        }, {
            "name": "periodenddate", "title": "Period End Date",
            "type": "date",
        }, {
            "name": "TR.CSRReportingScope", "title": "ESG Reporting Scope",
            "type": "number", "decimalChar": ".",
        }, {
            "name": "TR.ESGPeriodLastUpdateDate",
            "title": "ESG Period Last Update Date", "type": "datetime",
        }, {
            "name": "TR.CO2EmissionTotal",
            "title": "CO2 Equivalents Emission Total", "type": "number",
            "decimalChar": ".",
        }, {
            "name": "TR.WomenManagers", "title": "Women Managers",
            "type": "number", "decimalChar": ".",
        }, {
            "name": "TR.AvgTrainingHours", "title": "Average Training Hours",
            "type": "number", "decimalChar": ".",
        }]})
    # fmt: on
    session = StubSession(is_open=True, response=response)

    definition = esg.basic_overview.Definition(universe=["IBM.N"])
    response = definition.get_data(session)

    df = response.data.df
    assert is_datetime64_any_dtype(df["Period End Date"])



__all__ = (
    "close_session",
    "content",
    "dates_and_calendars",
    "delivery",
    "discovery",
    "eikon",
    "errors",
    "get_config",
    "get_data",
    "get_history",
    "load_config",
    "news",
    "open_pricing_stream",
    "open_session",
    "OpenState",
    "PricingStream",
    "session",
    "usage_collection",
    "_qpl",
)

"""
    refinitiv-data is a Python library to access Refinitiv Data Platform with Python.
"""

# _version.py file is generated by setuptools_scm on build
try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"

from ._config_functions import get_config, load_config
from ._open_state import OpenState
from . import delivery, session, content, errors, eikon, usage_collection, discovery
from ._fin_coder_layer.session import open_session, close_session
from ._fin_coder_layer.get_data import get_data
from ._fin_coder_layer.get_history import get_history
from ._fin_coder_layer.get_stream import PricingStream, open_pricing_stream
from ._fin_coder_layer import dates_and_calendars, news
from . import _qpl



import refinitiv.data as rd
from refinitiv.data._core.session.grant_password import GrantPassword
from refinitiv.data._core.session.http_service import get_http_limits


def test_session_event_code():
    import refinitiv.data as rd

    try:
        rd.session.EventCode
    except Exception as e:
        assert False, str(e)


def test_session_id():
    session_1 = rd.session.desktop.Definition(app_key="foo").get_session()
    session_2 = rd.session.platform.Definition(
        app_key="foo", grant=GrantPassword(username="username", password="password")
    ).get_session()
    session_3 = rd.session.desktop.Definition(app_key="foo").get_session()

    assert session_1.session_id is not None
    assert session_2.session_id == session_1.session_id + 1
    assert session_3.session_id == session_2.session_id + 1


def test_get_http_limits():
    # given
    max_connections = 21
    max_keepalive_connections = max_connections
    config = {
        "http.max-connections": max_connections,
        "http.max-keepalive-connections": max_keepalive_connections,
    }

    # when
    result = get_http_limits(config)

    # then
    assert result.max_connections == max_connections
    assert result.max_keepalive_connections == max_keepalive_connections



from enum import Enum, auto


class ContentType(Enum):
    CHAINS = auto()
    CONTRACTS = auto()
    CUSTOM_INSTRUMENTS_EVENTS = auto()
    CUSTOM_INSTRUMENTS_INSTRUMENTS = auto()
    CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES = auto()
    CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES = auto()
    CUSTOM_INSTRUMENTS_SEARCH = auto()
    DATA_GRID_RDP = auto()
    DATA_GRID_UDF = auto()
    DEFAULT = auto()
    DISCOVERY_LOOKUP = auto()
    DISCOVERY_METADATA = auto()
    DISCOVERY_SEARCH = auto()
    ESG_BASIC_OVERVIEW = auto()
    ESG_FULL_MEASURES = auto()
    ESG_FULL_SCORES = auto()
    ESG_STANDARD_MEASURES = auto()
    ESG_STANDARD_SCORES = auto()
    ESG_UNIVERSE = auto()
    ESTIMATES_VIEW_ACTUALS_ANNUAL = auto()
    ESTIMATES_VIEW_ACTUALS_INTERIM = auto()
    ESTIMATES_VIEW_ACTUALS_KPI_ANNUAL = auto()
    ESTIMATES_VIEW_ACTUALS_KPI_INTERIM = auto()
    ESTIMATES_VIEW_SUMMARY_ANNUAL = auto()
    ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_NON_PERIODIC_MEASURES = auto()
    ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_ANNUAL = auto()
    ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_INTERIM = auto()
    ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_RECOMMENDATIONS = auto()
    ESTIMATES_VIEW_SUMMARY_INTERIM = auto()
    ESTIMATES_VIEW_SUMMARY_KPI_ANNUAL = auto()
    ESTIMATES_VIEW_SUMMARY_KPI_HISTORICAL_SNAPSHOTS_KPI = auto()
    ESTIMATES_VIEW_SUMMARY_KPI_INTERIM = auto()
    ESTIMATES_VIEW_SUMMARY_NON_PERIODIC_MEASURES = auto()
    ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS = auto()
    FILINGS_RETRIEVAL = auto()
    FILINGS_SEARCH = auto()
    BOND_CURVE = auto()
    FORWARD_CURVE = auto()
    CROSS_CURRENCY_CURVES_CURVES = auto()
    CROSS_CURRENCY_CURVES_DEFINITIONS_CREATE = auto()
    CROSS_CURRENCY_CURVES_DEFINITIONS_DELETE = auto()
    CROSS_CURRENCY_CURVES_DEFINITIONS_GET = auto()
    CROSS_CURRENCY_CURVES_DEFINITIONS_UPDATE = auto()
    CROSS_CURRENCY_CURVES_DEFINITIONS_SEARCH = auto()
    CROSS_CURRENCY_CURVES_TRIANGULATE_DEFINITIONS_SEARCH = auto()
    HISTORICAL_PRICING_EVENTS = auto()
    HISTORICAL_PRICING_INTERDAY_SUMMARIES = auto()
    HISTORICAL_PRICING_INTRADAY_SUMMARIES = auto()
    NEWS_HEADLINES_RDP = auto()
    NEWS_HEADLINES_UDF = auto()
    NEWS_STORY_RDP = auto()
    NEWS_IMAGES = auto()
    NEWS_STORY_UDF = auto()
    NEWS_TOP_NEWS = auto()
    NEWS_TOP_NEWS_HEADLINES = auto()
    NONE = auto()
    OWNERSHIP_CONSOLIDATED_BREAKDOWN = auto()
    OWNERSHIP_CONSOLIDATED_CONCENTRATION = auto()
    OWNERSHIP_CONSOLIDATED_INVESTORS = auto()
    OWNERSHIP_CONSOLIDATED_RECENT_ACTIVITY = auto()
    OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT = auto()
    OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_REPORT = auto()
    OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION = auto()
    OWNERSHIP_FUND_BREAKDOWN = auto()
    OWNERSHIP_FUND_CONCENTRATION = auto()
    OWNERSHIP_FUND_HOLDINGS = auto()
    OWNERSHIP_FUND_INVESTORS = auto()
    OWNERSHIP_FUND_RECENT_ACTIVITY = auto()
    OWNERSHIP_FUND_SHAREHOLDERS_HISTORY_REPORT = auto()
    OWNERSHIP_FUND_SHAREHOLDERS_REPORT = auto()
    OWNERSHIP_FUND_TOP_N_CONCENTRATION = auto()
    OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT = auto()
    OWNERSHIP_INSIDER_TRANSACTION_REPORT = auto()
    OWNERSHIP_INVESTOR_HOLDINGS = auto()
    OWNERSHIP_ORG_INFO = auto()
    PRICING = auto()
    STREAMING_CHAINS = auto()
    STREAMING_CONTRACTS = auto()
    STREAMING_CUSTOM = auto()
    STREAMING_CUSTOM_INSTRUMENTS = auto()
    STREAMING_PRICING = auto()
    STREAMING_TRADING = auto()
    STREAMING_CONTRIB = auto()
    STREAMING_OFF_CONTRIB = auto()
    SURFACES = auto()
    ZC_CURVE_DEFINITIONS = auto()
    ZC_CURVES = auto()
    DATES_AND_CALENDARS_ADD_PERIODS = auto()
    DATES_AND_CALENDARS_HOLIDAYS = auto()
    DATES_AND_CALENDARS_COUNT_PERIODS = auto()
    DATES_AND_CALENDARS_DATE_SCHEDULE = auto()
    DATES_AND_CALENDARS_IS_WORKING_DAY = auto()



from enum import Enum, unique


@unique
class Rating(Enum):
    A = "A"
    A1 = "A1"
    A2 = "A2"
    A3 = "A3"
    AA = "AA"
    AAA = "AAA"
    AA_HIGH = "AAHigh"
    AA_LOW = "AALow"
    AA_MINUS = "AAMinus"
    AA_PLUS = "AAPlus"
    A_HIGH = "AHigh"
    A_LOW = "ALow"
    A_MINUS = "AMinus"
    A_PLUS = "APlus"
    AA1 = "Aa1"
    AA2 = "Aa2"
    AA3 = "Aa3"
    A_AA = "Aaa"
    B = "B"
    B1 = "B1"
    B2 = "B2"
    B3 = "B3"
    BB = "BB"
    BBB = "BBB"
    BBB_HIGH = "BBBHigh"
    BBB_LOW = "BBBLow"
    BBB_MINUS = "BBBMinus"
    BBB_PLUS = "BBBPlus"
    BB_HIGH = "BBHigh"
    BB_LOW = "BBLow"
    BB_MINUS = "BBMinus"
    BB_PLUS = "BBPlus"
    B_HIGH = "BHigh"
    B_LOW = "BLow"
    B_MINUS = "BMinus"
    B_PLUS = "BPlus"
    BA1 = "Ba1"
    BA2 = "Ba2"
    BA3 = "Ba3"
    BAA1 = "Baa1"
    BAA2 = "Baa2"
    BAA3 = "Baa3"
    C = "C"
    CC = "CC"
    CCC = "CCC"
    CCC_HIGH = "CCCHigh"
    CCC_LOW = "CCCLow"
    CCC_MINUS = "CCCMinus"
    CCC_PLUS = "CCCPlus"
    CC_HIGH = "CCHigh"
    CC_LOW = "CCLow"
    C_HIGH = "CHigh"
    C_LOW = "CLow"
    CA = "Ca"
    CAA1 = "Caa1"
    CAA2 = "Caa2"
    CAA3 = "Caa3"
    D = "D"
    DD = "DD"
    DDD = "DDD"
    F1 = "F1"
    F1_PLUS = "F1Plus"
    F2 = "F2"
    F3 = "F3"



# coding: utf8


from enum import Enum, unique


@unique
class InputVolatilityType(Enum):
    DEFAULT = "Default"
    LOG_NORMAL_VOLATILITY = "LogNormalVolatility"
    NORMALIZED_VOLATILITY = "NormalizedVolatility"



# coding: utf8

from enum import Enum, unique


@unique
class Seniority(Enum):
    JUNIOR_SUBORDINATED = "JuniorSubordinated"
    NONE = "None"
    PREFERENCE = "Preference"
    SECURED = "Secured"
    SENIOR_UNSECURED = "SeniorUnsecured"
    SUBORDINATED = "Subordinated"



import pytest

from refinitiv.data import _configure as configure


@pytest.mark.parametrize(
    "method_name",
    [
        "desktop_session",
        "desktop_base_uri",
        "desktop_platform_paths",
        "desktop_handshake_url",
        "desktop_endpoints",
        "platform_session",
        "platform_endpoints",
        "platform_base_uri",
        "platform_auth_uri",
        "platform_token_uri",
        "platform_auto_reconnect",
        "platform_server_mode",
        "platform_realtime_distribution_system",
    ],
)
def test_config_keys(method_name):
    # given
    testing_subst = "test"
    configure.reload()

    # when
    testing_method = getattr(configure._keys, method_name)
    key = testing_method(testing_subst)

    assert testing_subst in key



# coding: utf8

__all__ = ["MoneynessWeight"]

from ..._object_definition import ObjectDefinition


class MoneynessWeight(ObjectDefinition):
    def __init__(self, max_moneyness=None, min_moneyness=None, weight=None):
        super().__init__()
        self.max_moneyness = max_moneyness
        self.min_moneyness = min_moneyness
        self.weight = weight

    @property
    def max_moneyness(self):
        """
        :return: float
        """
        return self._get_parameter("maxMoneyness")

    @max_moneyness.setter
    def max_moneyness(self, value):
        self._set_parameter("maxMoneyness", value)

    @property
    def min_moneyness(self):
        """
        :return: float
        """
        return self._get_parameter("minMoneyness")

    @min_moneyness.setter
    def min_moneyness(self, value):
        self._set_parameter("minMoneyness", value)

    @property
    def weight(self):
        """
        :return: float
        """
        return self._get_parameter("weight")

    @weight.setter
    def weight(self, value):
        self._set_parameter("weight", value)



from datetime import datetime, timedelta

from refinitiv.data.delivery import rdp_stream

today = datetime.utcnow()
yesterday = today - timedelta(days=1)
yesterday = yesterday.isoformat() + "Z"
today = today.isoformat() + "Z"


def get_financial_contracts_stream(extended_params=None):
    stream = rdp_stream.Definition(
        service=None,
        api="streaming.quantitative-analytics.endpoints.financial-contracts",
        universe={
                "instrumentType": "FxCross",
                "instrumentDefinition": {
                    "instrumentTag": "USDAUD",
                    "fxCrossType": "FxSpot",
                    "fxCrossCode": "USDAUD",
                },
            },
        parameters=None,
        extended_params=extended_params,
        view=[
            "InstrumentTag",
            "FxSpot_BidMidAsk",
            "ErrorCode",
            "Ccy1SpotDate",
            "Ccy2SpotDate",
        ],
    ).get_stream()
    return stream


def get_benchmark_stream(extended_params=None):
    stream = rdp_stream.Definition(
        service=None,
        api="streaming.benchmark.endpoints.resource",
        universe=["TSLA.O", "AAPL.O", "AMZN.O"],
        view=None,
        parameters={
            "calculationType": "vwap",
            "responseType": {"type": "streaming", "frequency": "tick"},
            "timeField": "exchange",
            "startTime": {"absoluteTime": yesterday},
            "endTime": {"absoluteTime": today},
        },
        extended_params=extended_params,
    ).get_stream()
    return stream


def get_trading_analytics_stream(extended_params=None):
    stream = rdp_stream.Definition(
        service=None,
        universe=[],
        view=None,
        parameters={"universeType": "RIC"},
        api="streaming.trading-analytics.endpoints.redi",
        extended_params=extended_params,
    ).get_stream()
    return stream


def add_callbacks(stream, stream_log=None):
    if stream_log is None:
        stream_log = []

    stream.on_update(
        lambda update, item_stream: on_stream_update(
            update, item_stream, "update", stream_log
        )
    )
    stream.on_ack(
        lambda ack_msg, item_stream: on_stream_update(
            ack_msg, item_stream, "ack", stream_log
        )
    )
    stream.on_alarm(
        lambda alarm_msg, item_stream: on_stream_update(
            alarm_msg, item_stream, "alarm", stream_log
        )
    )
    stream.on_response(
        lambda response, item_stream: on_stream_update(
            response, item_stream, "response", stream_log
        )
    )


def on_stream_update(event, stream, event_type, stream_log):
    stream_log.append(event_type)
    message = f"{datetime.now()}: {event_type} - {event} received at"
    print(message)


def check_stream_view(stream, expected_view):
    assert stream._view == expected_view


def check_stream_universe(stream, expected_universe):
    assert stream._universe == expected_universe



__all__ = (
    "BuySell",
    "Definition",
    "FxCrossType",
    "FxLegType",
    "FxPoint",
    "FxSwapCalculationMethod",
    "ImpliedDepositDateConvention",
    "LegDefinition",
    "PriceSide",
    "PricingParameters",
)


from ._definition import Definition
from ._enums import (
    BuySell,
    FxCrossType,
    FxLegType,
    FxSwapCalculationMethod,
    ImpliedDepositDateConvention,
    PriceSide,
)
from ._fx_cross_leg_definition import LegDefinition
from ._fx_cross_pricing_parameters import PricingParameters
from ._models import FxPoint



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    FxShiftScenario,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ShiftDefinition,
)


def test_fx_shift_scenario():
    # given
    expected_dict = {"fxCurveShift": {}, "shiftTag": "string"}

    # when
    testing_obj = FxShiftScenario(
        fx_curve_shift=ShiftDefinition(),
        shift_tag="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.fx_curve_shift == ShiftDefinition()
    assert testing_obj.shift_tag == "string"



import pandas as pd
import pytest
from refinitiv.data.delivery._data._parsed_data import ParsedData

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._df_builder_factory import get_dfbuilder
from refinitiv.data.content.filings._errors import DownloadFileError
from refinitiv.data.content.filings._search_data_provider import (
    FilingsSearchRequestFactory,
    FilingsSearchValidator,
    FilingsSearchFile,
    FilingsSearchData,
)


def test_filings_search_request_factory_get_body_parameters_query_override():
    # given
    data = {"query": "query", "variables": "variables"}

    # when
    body = FilingsSearchRequestFactory().get_body_parameters(**data)

    # then
    assert body.get("query") == "query"


def test_filings_search_data_df():
    # given
    raw_data = {
        "data": {
            "FinancialFiling": [
                {
                    "FilingDocument": {
                        "Identifiers": [{"Dcn": "dcn"}],
                        "DocId": "doc_id",
                        "FinancialFilingId": 1,
                        "DocumentSummary": {"DocumentTitle": "title"},
                        "FilesMetaData": [
                            {"FileName": "filename", "MimeType": "mimetype"}
                        ],
                    }
                }
            ]
        }
    }
    expected_raw_data = [["title", "filename", "mimetype", "dcn", "doc_id", 1]]
    columns = [
        "DocumentTitle",
        "Filename",
        "MimeType",
        "Dcn",
        "DocId",
        "FinancialFilingId",
    ]
    expected_df = pd.DataFrame(data=expected_raw_data, columns=columns)
    df_builder = get_dfbuilder(ContentType.FILINGS_SEARCH)

    # when
    data = FilingsSearchData(raw_data, use_field=False, dfbuilder=df_builder)
    df = data.df

    # then
    assert df.compare(expected_df).empty


def test_filing_file_download_none_error():
    # when
    f = FilingsSearchFile("name")

    # then
    with pytest.raises(DownloadFileError) as err:
        f.download()

    assert (
        err.value.message
        == "Cannot download file. Missing one of Filename, DCN, DocID and Filing ID"
    )


async def test_filing_file_download_async_none_error():
    # when
    f = FilingsSearchFile("name")

    # then
    response = await f.download_async()

    assert (
        response.errors[0].message
        == "Cannot download file. Missing one of Filename, DCN, DocID and Filing ID"
    )


def test_filing_content_data_has_errors():
    # given
    raw_data = {
        "errors": [
            {
                "message": "Validation error",
                "locations": [{"line": 1, "column": 4}],
                "extensions": {"classification": "ValidationError"},
            }
        ],
        "data": "None",
    }

    # when
    data = ParsedData(status=None, raw_response=raw_data, content_data=raw_data)

    # then
    assert not FilingsSearchValidator.content_data_has_no_errors(data)



"""ADC patterns for fincoder layer and fundamental and reference content object."""

import re

# re for ADC fields like started with "TR." case is ignored
ADC_TR_PATTERN = re.compile(r"^tr\..+", re.I)

# re for ADC date fields ended witH ".date" case is ignored
ADC_DATE_PATTERN = re.compile(r"^.+\.date$", re.I)

# re for finding expressions inside ADC fields like
# "TR.F.TotRevPerShr(SDate=0,EDate=-2,Period=FY0,Frq=FY).date"
ADC_PARAM_IN_FIELDS = re.compile(r".*\(.+\).*")

# re for ADC functions in fields like AVAIL(, AVG(
# AVAIL(TR.GrossProfit(Period=LTM,Methodology=InterimSum))
ADC_FUNC_PATTERN_IN_FIELDS = re.compile(r"^[A-Z]+\(")

# re for finding column name include sub-string
# ..._DATE...int
# ...DATE...int
# ..._DT...int
# ...DT...int
# ...DAT...int
PRICING_DATETIME_PATTERN = re.compile(r".*(DATE|DT|DAT)\d*$")



from itertools import product

import pandas as pd
from ...._tools._dataframe import (
    convert_df_columns_to_datetime,
    convert_dtypes,
    convert_df_columns_to_datetime_by_idx,
)


def bond_curve_build_df(raw, **kwargs):
    """
    Parameters
    ----------
    raw : dict

    Returns
    -------
    DataFrame

    Examples
    -------
    >>> raw
    ... {
    ...     "data": [
    ...         {
    ...             "error": {
    ...                 "id": "b6f9797d-72c8-4baa-84eb-6a079fc40ec5/b6f9797d-72c8-4baa-84eb-6a079fc40ec5",
    ...                 "code": "QPS-Curves.6",
    ...                 "message": "Invalid input: curveDefinition is missing",
    ...             }
    ...         },
    ...         {
    ...             "curveTag": "test_curve",
    ...             "curveParameters": {
    ...                 "interestCalculationMethod": "Dcb_Actual_Actual",
    ...                 "priceSide": "Mid",
    ...                 "calendarAdjustment": "Calendar",
    ...                 "calendars": ["EMU_FI"],
    ...                 "compoundingType": "Compounded",
    ...                 "useConvexityAdjustment": True,
    ...                 "useSteps": False,
    ...                 "valuationDate": "2022-02-09",
    ...             },
    ...             "curveDefinition": {
    ...                 "availableTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "availableDiscountingTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "currency": "EUR",
    ...                 "mainConstituentAssetClass": "Swap",
    ...                 "riskType": "InterestRate",
    ...                 "indexName": "EURIBOR",
    ...                 "source": "Refinitiv",
    ...                 "name": "EUR EURIBOR Swap ZC Curve",
    ...                 "id": "9d619112-9ab3-45c9-b83c-eb04cbec382e",
    ...                 "discountingTenor": "OIS",
    ...                 "ignoreExistingDefinition": False,
    ...                 "owner": "Refinitiv",
    ...             },
    ...             "curvePoints": [
    ...                 {
    ...                     "endDate": "2021-02-01",
    ...                     "startDate": "2021-02-01",
    ...                     "discountFactor": 1.0,
    ...                     "ratePercent": 7.040811073443143,
    ...                     "tenor": "0D",
    ...                 },
    ...                 {
    ...                     "endDate": "2021-02-04",
    ...                     "startDate": "2021-02-01",
    ...                     "discountFactor": 0.999442450671571,
    ...                     "ratePercent": 7.040811073443143,
    ...                     "tenor": "1D",
    ...                 },
    ...             ],
    ...         },
    ...     ]
    ... }
    """
    datas = raw.get("data", [])
    datas = datas or []
    dfs = []
    for data in datas:

        error = data.get("error")
        if error:
            continue

        curve_points = data.get("curvePoints")

        for curve_point in curve_points:
            d = {}
            for key, value in curve_point.items():
                values = d.setdefault(key, [])
                values.append(value)

            df = pd.DataFrame(d)
            df = df.convert_dtypes()
            dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)
    df = convert_df_columns_to_datetime(df, "Date", utc=True, delete_tz=True)
    df = convert_dtypes(df)
    return df


def cross_currency_curves_curve_build_df(raw, **kwargs):
    """
    Parameters
    ----------
    raw : dict

    Returns
    -------
    DataFrame

    Examples
    -------
    >>> raw
    ... {
    ... 'data':
    ...    [
    ...        {
    ...            "error": {
    ...                "id": "b6f9797d-72c8-4baa-84eb-6a079fc40ec5/b6f9797d-72c8-4baa-84eb-6a079fc40ec5",
    ...                "code": "QPS-Curves.6",
    ...                "message": "Invalid input: curveDefinition is missing",
    ...            }
    ...        },
    ...        {
    ...            'curveDefinition': {
    ...                'baseCurrency': 'EUR', 'baseIndexName': 'ESTR',
    ...                'quotedCurrency': 'USD', 'quotedIndexName': 'SOFR',
    ...                'crossCurrencyDefinitions': [
    ...                    {
    ...                        'baseCurrency': 'EUR', 'baseIndexName': 'ESTR',
    ...                        'name': 'EUR ESTR/USD SOFR FxCross',
    ...                        'quotedCurrency': 'USD', 'quotedIndexName': 'SOFR',
    ...                        'source': 'Refinitiv', 'isNonDeliverable': False,
    ...                        'mainConstituentAssetClass': 'Swap',
    ...                        'riskType': 'CrossCurrency',
    ...                        'id': 'c9f2e9fb-b04b-4140-8377-8b7e47391486',
    ...                        'ignoreExistingDefinition': False
    ...                    }
    ...                ]
    ...            },
    ...            'curveParameters': {
    ...                'valuationDate': '2021-10-06',
    ...                'interpolationMode': 'Linear',
    ...                'extrapolationMode': 'Constant', 'turnAdjustments': {},
    ...                'ignorePivotCurrencyHolidays': False,
    ...                'useDelayedDataIfDenied': False,
    ...                'ignoreInvalidInstrument': True,
    ...                'marketDataLookBack': {'value': 10, 'unit': 'CalendarDay'}
    ...            },
    ...            'curve': {
    ...                'fxCrossScalingFactor': 1.0, 'fxSwapPointScalingFactor': 10000.0,
    ...                'curvePoints': [
    ...                    {
    ...                        'tenor': 'SPOT', 'startDate': '2021-10-08',
    ...                        'endDate': '2021-10-08',
    ...                        'swapPoint': {'bid': 0.0, 'ask': 0.0, 'mid': 0.0},
    ...                        'outright': {'bid': 1.1556, 'ask': 1.156, 'mid': 1.1558}
    ...                    },
    ...                    {
    ...                        'tenor': 'SN', 'startDate': '2021-10-08',
    ...                        'endDate': '2021-10-12',
    ...                        'instruments': [{'instrumentCode': 'EURSN='}],
    ...                        'swapPoint': {
    ...                            'bid': 0.8399999999997299,
    ...                            'ask': 0.8800000000008801,
    ...                            'mid': 0.860000000000305
    ...                        },
    ...                        'outright': {'bid': 1.155684, 'ask': 1.156088,
    ...                                     'mid': 1.155886}
    ...                    }
    ...                ]
    ...            }
    ...        }
    ...    ]
    ... }
    """
    datas = raw.get("data", [])
    datas = datas or []
    get_curve = (data["curve"] for data in datas if "curve" in data)
    columns_have_level_1 = ("swapPoint", "outright")
    level_1 = ("bid", "ask", "mid")

    curve = next(get_curve, {})
    curve_points = curve.get("curvePoints", [])
    curve_point = max(curve_points, key=lambda x: len(x))
    columns = [key for key in curve_point]
    data_df = _create_data_for_df(curve_points, columns, columns_have_level_1, level_1)

    allcolumns = []
    for name in columns:
        if name in columns_have_level_1:
            allcolumns.extend(list(product([name], level_1)))
        else:
            allcolumns.append((name, ""))

    columns_date_idxs = [
        index for index, value in enumerate(allcolumns) if "date" in value[0].lower()
    ]
    columns = pd.MultiIndex.from_tuples(allcolumns)
    df = pd.DataFrame(data_df, columns=columns)
    df = convert_df_columns_to_datetime_by_idx(
        df, columns_date_idxs, utc=True, delete_tz=True
    )
    df = convert_dtypes(df)
    return df


def _create_data_for_df(curve_points, columns, columns_have_level_1, columns_level_1):
    data_df = []

    for curve_point in curve_points:
        row_data = []
        for name in columns:
            value = curve_point.get(name, pd.NA)
            if name == "instruments" and not pd.isna(value):
                value = [v["instrumentCode"] for v in value if "instrumentCode" in v]
                value = value.pop() if len(value) == 1 else value
            if name in columns_have_level_1:
                value = [value.get(i, pd.NA) for i in columns_level_1]
                row_data.extend(value)
            else:
                row_data.append(value)
        data_df.append(row_data)

    return data_df



import itertools
import threading
from concurrent.futures import ThreadPoolExecutor, wait
from typing import (
    TYPE_CHECKING,
    Union,
    Iterable,
    Callable,
    Dict,
    KeysView,
    ValuesView,
    Tuple,
    ItemsView,
    Optional,
)

import pandas as pd

from ._universe_stream import _UniverseStream
from .._core.session import Session, DesktopSession
from .._errors import ItemWasNotRequested
from .._tools import (
    universe_arg_parser,
    fields_arg_parser,
    cached_property,
    try_copy_to_list,
)
from .._types import Strings, OptBool, OptStr
from ..delivery._stream import (
    StreamStateManager,
    OMMStreamListener,
    StreamStateEvent,
    StreamEvent,
)
from ..delivery._stream.base_stream import StreamOpenWithUpdatesMixin
from ..delivery.omm_stream import ContribResponse

if TYPE_CHECKING:
    from ..delivery.omm_stream import ContribType

_id_iterator = itertools.count()


def build_df(
    universe: Strings, fields: Strings, values_by_field: dict, convert: bool
) -> pd.DataFrame:
    data = []
    for inst_name in universe:
        items = []

        for values in values_by_field.values():
            item = values.pop(0)

            if item is None:
                item = pd.NA

            items.append(item)

        data.append((inst_name, *items))

    df = pd.DataFrame(data=data, columns=("Instrument", *fields))

    if convert:
        df = df.convert_dtypes()

    return df


def validate(name, input_values, requested_values):
    input_values = set(input_values)
    requested_values = set(requested_values)
    not_requested = input_values - requested_values
    has_not_requested = bool(not_requested)

    if has_not_requested:
        raise ItemWasNotRequested(name, not_requested, requested_values)


def get_available_fields(value_by_fields_by_inst_names):
    available_fields = []
    for value_by_fields in value_by_fields_by_inst_names.values():
        fields = value_by_fields.keys()
        available_fields.extend(fields)
    return list(dict.fromkeys(available_fields))


class UniverseStreamFacade(StreamOpenWithUpdatesMixin):
    def __init__(self, stream: _UniverseStream):
        self._stream: _UniverseStream = stream

    @property
    def id(self) -> int:
        return self._stream.id

    @property
    def code(self):
        return self._stream.code

    @property
    def message(self):
        return self._stream.message

    @property
    def name(self):
        return self._stream.name

    @property
    def service(self):
        return self._stream.service

    @property
    def fields(self):
        return self._stream.fields

    @property
    def status(self):
        return self._stream.status

    @property
    def is_ok(self):
        return self._stream.is_ok

    def __enter__(self):
        return self._stream.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._stream.__exit__(exc_type, exc_val, exc_tb)

    def get_field_value(self, field):
        return self._stream.get_field_value(field)

    def get_fields(self, fields=None) -> dict:
        return self._stream.get_fields(fields)

    def keys(self):
        return self._stream.keys()

    def values(self):
        return self._stream.values()

    def items(self):
        return self._stream.items()

    def __iter__(self):
        return self._stream.__iter__()

    def __getitem__(self, field):
        try:
            item = self._stream.__getitem__(field)
        except KeyError:
            item = None
        return item

    def __len__(self):
        return self._stream.__len__()

    def __repr__(self):
        return self._stream.__repr__()

    def __str__(self):
        return self._stream.__str__()


class _UniverseStreams(StreamStateManager, OMMStreamListener):
    def __init__(
        self,
        content_type,
        universe: Union[str, Iterable[str]],
        session: "Session",
        item_facade_class=UniverseStreamFacade,
        fields: Union[str, list] = None,
        service: str = None,
        on_refresh: Callable = None,
        on_status: Callable = None,
        on_update: Callable = None,
        on_complete: Callable = None,
        on_ack: Callable = None,
        api: OptStr = None,
        extended_params: dict = None,
    ) -> None:
        self._session = session

        StreamStateManager.__init__(self, logger=self.session.logger())
        OMMStreamListener.__init__(
            self,
            logger=self.session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
            on_ack=on_ack,
        )

        self._universe: Strings = universe_arg_parser.get_list(universe)
        self.fields: Strings = fields_arg_parser.get_list(fields or [])
        self._service = service
        self._api = api
        self._extended_params = extended_params
        self._id = next(_id_iterator)
        self._classname: str = (
            f"[{self.__class__.__name__} id={self._id} universe={universe}]"
        )
        self._completed = set()
        self._content_type = content_type
        self._item_facade_class = item_facade_class
        self._lock_insts_fields = threading.Lock()

    @property
    def session(self) -> "Session":
        return self._session

    @session.setter
    def session(self, session: "Session"):
        if self._session != session and not self.is_open:
            self._session = session
            OMMStreamListener.init_logger(self, self._session.logger())

            for stream in self.values():
                stream.session = self._session

    @property
    def universe(self):
        return self._universe

    def _create_stream_by_name(self, name):
        return _UniverseStream(
            content_type=self._content_type,
            session=self._session,
            name=name,
            fields=self.fields,
            service=self._service,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
            on_ack=self._on_stream_ack,
            api=self._api,
            extended_params=self._extended_params,
            parent_id=self._id,
        )

    @cached_property
    def _stream_by_name(self) -> Dict[str, _UniverseStream]:
        retval = {}
        for name in self.universe:
            stream = self._create_stream_by_name(name)
            stream.on(StreamStateEvent.CLOSED, self._on_stream_close)
            retval[name] = stream
        return retval

    def keys(self) -> KeysView[str]:
        return self._stream_by_name.keys()

    def values(self) -> ValuesView[_UniverseStream]:
        return self._stream_by_name.values()

    def items(self) -> ItemsView[str, _UniverseStream]:
        return self._stream_by_name.items()

    def __iter__(self):
        return StreamIterator(self)

    def __getitem__(self, name) -> Union[_UniverseStream, dict, UniverseStreamFacade]:
        stream = self._stream_by_name.get(name, {})

        if stream == {}:
            self._warning(f"'{name}' not in {self._classname} universe")
            return stream

        if hasattr(stream, "_wrapper"):
            wrapper = stream._wrapper
        else:
            wrapper = self._item_facade_class(stream)
            stream._wrapper = wrapper

        return wrapper

    def __len__(self):
        return len(self._stream_by_name)

    def _get_value_by_fields_by_inst_names(
        self, universe: "Strings", fields: "Strings"
    ) -> dict:
        retval = {}
        for name in universe:
            stream = self._stream_by_name.get(name)

            values_by_fields = stream.get_fields(fields)
            if values_by_fields:
                retval[name] = values_by_fields
        return retval

    def get_snapshot(
        self,
        universe: Optional[Union[str, Strings]] = None,
        fields: Optional[Union[str, Strings]] = None,
        convert: OptBool = True,
    ) -> pd.DataFrame:
        """
        Returns a Dataframe filled with snapshot values
        for a list of instrument names and a list of fields.

        Parameters
        ----------
        universe: list of strings
            List of instruments to request snapshot data on.

        fields: str or list of strings
            List of fields to request.

        convert: boolean
            If True, force numeric conversion for all values.

        Returns
        -------
            pandas.DataFrame

            pandas.DataFrame content:
                - columns : instrument and field names
                - rows : instrument name and field values

        Raises
        ------
            Exception
                If request fails or if server returns an error

            ValueError
                If a parameter type or value is wrong

        """

        if universe:
            universe = try_copy_to_list(universe)
            universe = universe_arg_parser.get_list(universe)
            try:
                validate("Instrument", universe, self.universe)
            except ItemWasNotRequested as e:
                self._error(e)
                return pd.DataFrame()
        else:
            universe = self.universe

        value_by_fields_by_inst_names: dict = self._get_value_by_fields_by_inst_names(
            universe, fields
        )

        if fields:
            fields = try_copy_to_list(fields)
            fields = fields_arg_parser.get_unique(fields or [])

            try:
                self.fields and validate("Field", fields, self.fields)
            except ItemWasNotRequested as e:
                self._error(e)
                return pd.DataFrame()
        else:
            fields = get_available_fields(value_by_fields_by_inst_names)

        values_by_field = {
            field: [
                value_by_fields_by_inst_names.get(name, {}).get(field)
                for name in universe
            ]
            for field in fields
        }

        return build_df(universe, fields, values_by_field, convert)

    def contribute(
        self,
        name: str,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> "ContribResponse":
        self._debug(f"{self._classname} contribute on {name}")
        stream = self._stream_by_name.get(name)
        if stream:
            return stream.contribute(fields, contrib_type, post_user_info)
        else:
            self._error(f"Can't contribute to unsubscribed item {name}")
            return ContribResponse({})

    async def contribute_async(
        self,
        name: str,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> "ContribResponse":
        self._debug(f"{self._classname} contribute on {name}")
        stream = self._stream_by_name.get(name)
        if stream:
            return await stream.contribute_async(
                fields=fields, contrib_type=contrib_type, post_user_info=post_user_info
            )
        else:
            self._error(f"Can't contribute to unsubscribed item {name}")
            return ContribResponse({})

    def _do_open(self, *args, with_updates=True) -> None:
        self._debug(f"{self._classname} open streaming on {self.universe}")

        if not self.values():
            raise ValueError("No instrument to subscribe")

        self._completed.clear()

        with ThreadPoolExecutor(
            thread_name_prefix="OpenUniverseStreams-Thread"
        ) as executor:
            futures = [
                executor.submit(stream.open, with_updates=with_updates)
                for stream in self.values()
            ]
            wait(futures)
            for fut in futures:
                exception = fut.exception()
                if exception:
                    raise exception

        self._debug(f"{self._classname} streaming on {self.universe} is open")

    def _do_close(self, *args, **kwargs) -> None:
        self._debug(f"{self._classname} close streaming on {str(self.universe)}")

        with ThreadPoolExecutor(
            thread_name_prefix="CloseUniverseStreams-Thread"
        ) as executor:
            futures = [executor.submit(stream.close) for stream in self.values()]
            wait(futures)
            for fut in futures:
                exception = fut.exception()
                if exception:
                    raise exception

    def _on_stream_close(self, stream) -> None:
        self._debug(f"{self._classname} streaming closed name={stream.name}")
        self.close()

    def _do_on_stream_refresh(self, stream, *args) -> Tuple:
        return self._do_on_stream(stream, *args)

    def _on_stream_status(self, stream, status, *args) -> None:
        self._debug(f"{self._classname} on_status {args}")

        self._emitter.emit(StreamEvent.STATUS, self, stream.name, status)

        if stream.is_closed:
            self.dispatch_complete(stream)

    def _do_on_stream_update(self, stream, *args) -> Tuple:
        return self._do_on_stream(stream, *args)

    def _do_on_stream(self, stream, fields, *args) -> Tuple:
        return stream.name, fields

    def _on_stream_ack(self, stream, ack_msg, *args) -> None:
        self._debug(f"{self._classname} on_ack {args}")

        self._emitter.emit(StreamEvent.ACK, self, stream.name, ack_msg)

    def _on_stream_complete(self, stream, *args) -> None:
        self._debug(f"{self._classname} on_complete {args}")

        name = stream.name
        if name in self._completed:
            return

        self._completed.update([name])
        if self._completed == set(self.universe):
            # received complete event from all streams all, emit global complete
            self._emitter.emit(StreamEvent.COMPLETE, self)

    def _do_on_stream_error(self, stream, error, *args) -> Tuple:
        return stream.name, error

    def _update_fields_desktop_session(self):
        for name in self._stream_by_name:
            stream = self._stream_by_name[name]
            # main stream should not close
            stream.off(StreamStateEvent.CLOSED, self._on_stream_close)
            stream.close()
            stream.on(StreamStateEvent.CLOSED, self._on_stream_close)
            stream.open()

    def _update_fields_platform_session(self):
        for name in self._stream_by_name:
            self._stream_by_name[name].send_open_message()

    def _remove_fields_from_record(self, fields):
        for name in self._stream_by_name:
            self._stream_by_name[name].remove_fields_from_record(fields)

    def add_fields(self, fields) -> None:
        with self._lock_insts_fields:

            fields = fields_arg_parser.get_list(fields)

            exists_fields = set(fields) & set(self.fields)
            if exists_fields:
                self._error(f"{exists_fields} already in fields list")

            fields = [
                i for i in fields if i not in self.fields
            ]  # universe should be unique
            self.fields.extend(fields)
            if self.is_open:
                if isinstance(self._session, DesktopSession):
                    self._update_fields_desktop_session()
                else:
                    self._update_fields_platform_session()

    def remove_fields(self, fields) -> None:
        with self._lock_insts_fields:

            fields = fields_arg_parser.get_list(fields)

            not_exists_fields = set(fields) - set(self.fields)
            if not_exists_fields:
                self._error(f"{not_exists_fields} not in fields list")

            fields = [i for i in fields if i in self.fields]

            for i in fields:
                self.fields.remove(i)

            if self.is_open:
                if isinstance(self._session, DesktopSession):
                    self._update_fields_desktop_session()
                else:
                    self._update_fields_platform_session()
                self._remove_fields_from_record(fields)

    def _add_instrument(self, instrument):
        if instrument in self._universe:
            self._error(f"{instrument} already in universe list")
            return

        new_stream = self._create_stream_by_name(instrument)
        new_stream.on(StreamStateEvent.CLOSED, self._on_stream_close)
        if not self.is_unopened:
            self._stream_by_name[instrument] = new_stream

        if self.is_open:
            new_stream.off(StreamEvent.COMPLETE, self._on_stream_complete)
            new_stream.open()
        self._universe.append(instrument)

    def _remove_instrument(self, instrument):
        if instrument not in self._universe:
            self._error(f"{instrument} not in universe list")
            return

        self._universe.remove(instrument)

        if self.is_unopened:
            return

        if self.is_open:
            # main stream should not close
            stream = self._stream_by_name[instrument]
            stream.off(StreamStateEvent.CLOSED, self._on_stream_close)
            stream.close()
        del self._stream_by_name[instrument]

    def add_instruments(self, instruments) -> None:
        with self._lock_insts_fields:

            instruments = universe_arg_parser.get_list(instruments)

            if self.is_unopened:
                # universe update is enough, cache will create when stream open
                self._universe.extend(instruments)
                return

            for instrument in instruments:
                self._add_instrument(instrument)

    def remove_instruments(self, instruments) -> None:
        with self._lock_insts_fields:

            instruments = universe_arg_parser.get_list(instruments)

            if not self._universe:
                self._error("nothing to delete")
                return

            for instrument in instruments:
                self._remove_instrument(instrument)


class StreamIterator:
    def __init__(self, stream: _UniverseStreams):
        self._streaming_prices = stream
        self._index = 0

    def __next__(self):
        if self._index < len(self._streaming_prices._universe):
            result = self._streaming_prices[
                self._streaming_prices._universe[self._index]
            ]
            self._index += 1
            return result
        raise StopIteration()
